import pytest
from hpcflow.api import (
    InputSourceType,
    Parameter,
    SchemaInput,
    SchemaOutput,
    TaskSchema,
    TaskObjective,
    TaskSourceType,
    Task,
    InputValue,
    InputSource,
    InputSourceMode,
)
from hpcflow.sdk.core.errors import (
    TaskTemplateInvalidNesting,
    TaskTemplateMultipleInputValues,
    TaskTemplateMultipleSchemaObjectives,
    TaskTemplateUnexpectedInput,
)


@pytest.fixture
def param_p1():
    return Parameter("p1")


@pytest.fixture
def param_p2():
    return Parameter("p2")


@pytest.fixture
def param_p3():
    return Parameter("p3")


def test_task_expected_input_source_mode_no_sources(param_p1):
    s1 = TaskSchema("ts1", actions=[], inputs=[SchemaInput(param_p1)])
    t1 = Task(
        schemas=s1,
        inputs=[InputValue(param_p1, value=101)],
    )
    assert t1.input_source_mode == InputSourceMode.AUTO


def test_task_expected_input_source_mode_with_sources(param_p1):
    s1 = TaskSchema("ts1", actions=[], inputs=[SchemaInput(param_p1)])
    t1 = Task(
        schemas=s1,
        inputs=[InputValue(param_p1, value=101)],
        input_sources=[InputSource.local()],
    )
    assert t1.input_source_mode == InputSourceMode.MANUAL


def test_task_get_available_task_input_sources_expected_return_first_task_local_value(
    param_p1,
):

    s1 = TaskSchema("ts1", actions=[], inputs=[SchemaInput(param_p1)])
    t1 = Task(schemas=s1, inputs=[InputValue(param_p1, value=101)])

    available = t1.get_available_task_input_sources()
    available_exp = {"p1": [InputSource(source_type=InputSourceType.LOCAL)]}

    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_first_task_default_value(
    param_p1,
):

    s1 = TaskSchema("ts1", actions=[], inputs=[SchemaInput(param_p1, default_value=101)])

    t1 = Task(schemas=s1)

    available = t1.get_available_task_input_sources()
    available_exp = {"p1": [InputSource(source_type=InputSourceType.DEFAULT)]}

    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_one_output(
    param_p1, param_p2
):

    s1 = TaskSchema(
        "ts1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2)],
    )
    s2 = TaskSchema("ts2", actions=[], inputs=[SchemaInput(param_p2)])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2)

    t1._insert_ID = 0

    available = t2.get_available_task_input_sources([t1])
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            )
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_one_output_with_default(
    param_p1, param_p2
):

    s1 = TaskSchema(
        "ts1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2)],
    )
    s2 = TaskSchema("ts2", actions=[], inputs=[SchemaInput(param_p2, default_value=2002)])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2)

    t1._insert_ID = 0

    available = t2.get_available_task_input_sources([t1])
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            ),
            InputSource(source_type=InputSourceType.DEFAULT),
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_one_output_with_local(
    param_p1, param_p2
):

    s1 = TaskSchema(
        "ts1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2)],
    )
    s2 = TaskSchema("ts2", actions=[], inputs=[SchemaInput(param_p2)])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2, inputs=[InputValue(param_p2, value=202)])

    t1._insert_ID = 0

    available = t2.get_available_task_input_sources([t1])
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            ),
            InputSource(source_type=InputSourceType.LOCAL),
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_one_output_with_default_and_local(
    param_p1, param_p2
):

    s1 = TaskSchema(
        "ts1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2)],
    )
    s2 = TaskSchema("ts2", actions=[], inputs=[SchemaInput(param_p2, default_value=2002)])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2, inputs=[InputValue(param_p2, value=202)])

    t1._insert_ID = 0

    available = t2.get_available_task_input_sources([t1])
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            ),
            InputSource(source_type=InputSourceType.LOCAL),
            InputSource(source_type=InputSourceType.DEFAULT),
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_two_outputs(
    param_p1, param_p2, param_p3
):
    s1 = TaskSchema(
        "ts1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2), SchemaOutput(param_p3)],
    )
    s2 = TaskSchema(
        "ts2",
        actions=[],
        inputs=[SchemaInput(param_p2)],
        outputs=[SchemaOutput(param_p3), SchemaOutput("p4")],
    )
    s3 = TaskSchema("ts3", actions=[], inputs=[SchemaInput(param_p3)])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2)
    t3 = Task(schemas=s3)

    t1._insert_ID = 0
    t2._insert_ID = 1

    available = t3.get_available_task_input_sources([t1, t2])
    available_exp = {
        "p3": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            ),
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=1,
                task_source_type=TaskSourceType.OUTPUT,
            ),
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_two_params_one_output(
    param_p1,
    param_p2,
    param_p3,
):

    s1 = TaskSchema(
        "ts1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2), SchemaOutput(param_p3)],
    )
    s2 = TaskSchema(
        "ts2",
        actions=[],
        inputs=[SchemaInput(param_p2), SchemaInput(param_p3)],
    )

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2)

    t1._insert_ID = 0

    available = t2.get_available_task_input_sources([t1])
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            )
        ],
        "p3": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            )
        ],
    }
    assert available == available_exp


def test_get_task_unique_names_two_tasks_no_repeats():
    s1 = TaskSchema("ts1", actions=[])
    s2 = TaskSchema("ts2", actions=[])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2)

    assert Task.get_task_unique_names([t1, t2]) == ["ts1", "ts2"]


def test_get_task_unique_names_two_tasks_with_repeat():

    s1 = TaskSchema("ts1", actions=[])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s1)

    assert Task.get_task_unique_names([t1, t2]) == ["ts1_1", "ts1_2"]


def test_raise_on_multiple_schema_objectives():

    s1 = TaskSchema("t1", actions=[])
    s2 = TaskSchema("t2", actions=[])
    with pytest.raises(TaskTemplateMultipleSchemaObjectives):
        Task(schemas=[s1, s2])


def test_raise_on_unexpected_inputs(param_p1, param_p2):

    s1 = TaskSchema("t1", actions=[], inputs=[param_p1])

    with pytest.raises(TaskTemplateUnexpectedInput):
        Task(
            schemas=s1,
            inputs=[
                InputValue(param_p1, value=101),
                InputValue(param_p2, value=4),
            ],
        )


def test_raise_on_multiple_input_values(param_p1):

    s1 = TaskSchema("s1", inputs=[param_p1], actions=[])

    with pytest.raises(TaskTemplateMultipleInputValues):
        Task(
            schemas=s1,
            inputs=[
                InputValue(param_p1, value=101),
                InputValue(param_p1, value=7),
            ],
        )


def test_expected_return_defined_and_undefined_input_types(param_p1, param_p2):

    s1 = TaskSchema("s1", inputs=[param_p1, param_p2], actions=[])
    t1 = Task(schemas=s1, inputs=[InputValue(param_p1, value=101)])

    assert t1.defined_input_types == {param_p1.typ} and t1.undefined_input_types == {
        param_p2.typ
    }


def test_expected_return_all_schema_input_types_single_schema(param_p1, param_p2):

    s1 = TaskSchema("t1", actions=[], inputs=[param_p1, param_p2])
    t1 = Task(schemas=s1)

    assert t1.all_schema_input_types == {param_p1.typ, param_p2.typ}


def test_expected_return_all_schema_input_types_multiple_schemas(
    param_p1, param_p2, param_p3
):

    s1 = TaskSchema("t1", actions=[], inputs=[param_p1, param_p2])
    s2 = TaskSchema("t1", actions=[], inputs=[param_p1, param_p3])

    t1 = Task(schemas=[s1, s2])

    assert t1.all_schema_input_types == {param_p1.typ, param_p2.typ, param_p3.typ}


def test_expected_name_single_schema():
    s1 = TaskSchema("t1", actions=[])
    t1 = Task(schemas=[s1])
    assert t1.name == "t1"


def test_expected_name_single_schema_with_method():
    s1 = TaskSchema("t1", method="m1", actions=[])
    t1 = Task(schemas=s1)
    assert t1.name == "t1_m1"


def test_expected_name_single_schema_with_implementation():
    s1 = TaskSchema("t1", implementation="i1", actions=[])
    t1 = Task(schemas=s1)
    assert t1.name == "t1_i1"


def test_expected_name_single_schema_with_method_and_implementation():
    s1 = TaskSchema("t1", method="m1", implementation="i1", actions=[])
    t1 = Task(schemas=s1)
    assert t1.name == "t1_m1_i1"


def test_expected_name_multiple_schemas():
    s1 = TaskSchema("t1", actions=[])
    s2 = TaskSchema("t1", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1"


def test_expected_name_two_schemas_first_with_method():
    s1 = TaskSchema("t1", method="m1", actions=[])
    s2 = TaskSchema("t1", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1_m1"


def test_expected_name_two_schemas_first_with_method_and_implementation():
    s1 = TaskSchema("t1", method="m1", implementation="i1", actions=[])
    s2 = TaskSchema("t1", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1_m1_i1"


def test_expected_name_two_schemas_both_with_method():
    s1 = TaskSchema("t1", method="m1", actions=[])
    s2 = TaskSchema("t1", method="m2", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1_m1_and_m2"


def test_expected_name_two_schemas_first_with_method_second_with_implementation():
    s1 = TaskSchema("t1", method="m1", actions=[])
    s2 = TaskSchema("t1", implementation="i2", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1_m1_and_i2"


def test_expected_name_two_schemas_first_with_implementation_second_with_method():
    s1 = TaskSchema("t1", implementation="i1", actions=[])
    s2 = TaskSchema("t1", method="m2", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1_i1_and_m2"


def test_expected_name_two_schemas_both_with_method_and_implementation():
    s1 = TaskSchema("t1", method="m1", implementation="i1", actions=[])
    s2 = TaskSchema("t1", method="m2", implementation="i2", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1_m1_i1_and_m2_i2"


def test_raise_on_negative_nesting_order(param_p1):
    s1 = TaskSchema("t1", inputs=[param_p1], actions=[])
    with pytest.raises(TaskTemplateInvalidNesting):
        Task(schemas=s1, nesting_order={"p1": -1})


# TODO: test resolution of elements and with raise MissingInputs



from __future__ import annotations
from typing import Dict, List, Optional, Tuple, Union

from .json_like import ChildObjectSpec, JSONLike
from .command_files import FileSpec
from .element import ElementFilter, ElementGroup
from .errors import (
    MissingInputs,
    TaskTemplateInvalidNesting,
    TaskTemplateMultipleInputValues,
    TaskTemplateMultipleSchemaObjectives,
    TaskTemplateUnexpectedInput,
)
from .object_list import GroupList, ResourceList
from .parameters import (
    InputSource,
    InputSourceMode,
    InputSourceType,
    InputValue,
    ParameterPath,
    ResourceSpec,
    SchemaInput,
    SchemaOutput,
    TaskSourceType,
    ValuePerturbation,
    ValueSequence,
)
from .utils import get_duplicate_items, get_item_repeat_index


class Task(JSONLike):
    """Parametrisation of an isolated task for which a subset of input values are given
    "locally". The remaining input values are expected to be satisfied by other
    tasks/imports in the workflow."""

    # __slots__ = (
    #     "_schemas",
    #     "_repeats",
    #     "_resources",
    #     "_inputs",
    #     "_input_files",
    #     "_input_file_generator_sources",
    #     "_output_file_parser_sources",
    #     "_perturbations",
    #     "_sequences",
    #     "_input_sources",
    #     "_input_source_mode",
    #     "_nesting_order",
    #     "_groups",
    #     "_name",
    #     "_defined_input_types",  # assigned in _validate()
    #     "workflow_template",
    #     "_insert_ID",
    #     "_dir_name",
    # )

    _child_objects = (
        ChildObjectSpec(
            name="schemas",
            class_name="TaskSchema",
            is_multiple=True,
            shared_data_name="task_schemas",
            shared_data_primary_key="name",
        ),
        ChildObjectSpec(
            name="inputs",
            class_name="InputValue",
            is_multiple=True,
            dict_key_attr="parameter",
            dict_val_attr="value",
            parent_ref="_task",
        ),
        ChildObjectSpec(
            name="resources",
            class_name="ResourceList",
            parent_ref="_task",
        ),
        ChildObjectSpec(
            name="sequences",
            class_name="ValueSequence",
            is_multiple=True,
        ),
        ChildObjectSpec(
            name="input_sources",
            class_name="InputSource",
            is_multiple=True,
            is_dict_values=True,
            is_dict_values_ensure_list=True,
        ),
        ChildObjectSpec(
            name="input_source_mode",
            class_name="InputSourceMode",
            is_enum=True,
        ),
    )

    def __init__(
        self,
        schemas: Union[TaskSchema, str, List[TaskSchema], List[str]],
        repeats: Optional[Union[int, List[int]]] = 1,
        resources: Optional[Dict[str, Dict]] = None,
        inputs: Optional[List[InputValue]] = None,
        input_files: Optional[List[FileSpec]] = None,
        input_file_generator_sources: Optional[List] = None,
        output_file_parser_sources: Optional[List] = None,
        perturbations: Optional[List[ValuePerturbation]] = None,
        sequences: Optional[List[ValueSequence]] = None,
        input_sources: Optional[Dict[str, InputSource]] = None,
        input_source_mode: Optional[Union[str, InputSourceType]] = None,
        nesting_order: Optional[List] = None,
        groups: Optional[List[ElementGroup]] = None,
    ):
        # TODO: modify from_JSON_like(?) so "internal" attributes are not in init

        """
        Parameters
        ----------
        schema
            A (list of) `TaskSchema` object(s) and/or a (list of) strings that are task
            schema names that uniquely identify a task schema. If strings are provided,
            the `TaskSchema` object will be fetched from the known task schemas loaded by
            the app configuration.

        """

        # TODO: allow init via specifying objective and/or method and/or implementation
        # (lists of) strs e.g.: Task(
        #   objective='simulate_VE_loading',
        #   method=['CP_FFT', 'taylor'],
        #   implementation=['damask', 'damask']
        # )
        # where method and impl must be single strings of lists of the same length
        # and method/impl are optional/required only if necessary to disambiguate
        #
        # this would be like Task(schemas=[
        #   'simulate_VE_loading_CP_FFT_damask',
        #   'simulate_VE_loading_taylor_damask'
        # ])

        if not isinstance(schemas, list):
            schemas = [schemas]

        _schemas = []
        for i in schemas:
            if isinstance(i, str):
                try:
                    i = self.app.TaskSchema.get_by_key(i)
                except KeyError:
                    raise KeyError(f"TaskSchema {i!r} not found.")
            elif not isinstance(i, self.app.TaskSchema):
                raise TypeError(f"Not a TaskSchema object: {i!r}")
            _schemas.append(i)

        if isinstance(resources, dict):
            resources = self.app.ResourceList.from_json_like(resources)
        elif not resources:
            resources = self.app.ResourceList([self.app.ResourceSpec()])

        self._schemas = _schemas
        self._repeats = repeats
        self._resources = resources
        self._inputs = inputs or []
        self._input_files = input_files or []
        self._input_file_generator_sources = input_file_generator_sources or []
        self._output_file_parser_sources = output_file_parser_sources or []
        self._perturbations = perturbations or []
        self._sequences = sequences or []
        self._input_sources = input_sources or {}
        self._input_source_mode = input_source_mode or (
            InputSourceMode.MANUAL if input_sources else InputSourceMode.AUTO
        )
        self._nesting_order = nesting_order or {}
        self._groups = self.app.GroupList(groups or [])

        self._set_parent_refs()

        self._validate()
        self._name = self._get_name()

        self.workflow_template = None  # assigned by parent WorkflowTemplate
        self._insert_ID = None
        self._dir_name = None

    @classmethod
    def _json_like_constructor(cls, json_like):
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""
        insert_ID = json_like.pop("insert_ID", None)
        dir_name = json_like.pop("dir_name", None)
        obj = cls(**json_like)
        obj._insert_ID = insert_ID
        obj._dir_name = dir_name
        return obj

    def __repr__(self):
        return f"{self.__class__.__name__}(" f"name={self.name!r}" f")"

    def to_dict(self):
        out = super().to_dict()
        return {
            k.lstrip("_"): v
            for k, v in out.items()
            if k not in ["_name", "_defined_input_types"]
        }

    def _validate(self):

        # TODO: check a nesting order specified for each sequence?

        names = set(i.objective.name for i in self.schemas)
        if len(names) > 1:
            raise TaskTemplateMultipleSchemaObjectives(
                f"All task schemas used within a task must have the same "
                f"objective, but found multiple objectives: {list(names)!r}"
            )

        input_types = [i.parameter.typ for i in self.get_non_sub_parameter_input_values()]
        dup_params = get_duplicate_items(input_types)
        if dup_params:
            raise TaskTemplateMultipleInputValues(
                f"The following parameters are associated with multiple input value "
                f"definitions: {dup_params!r}."
            )

        unexpected_types = set(input_types) - self.all_schema_input_types
        if unexpected_types:
            raise TaskTemplateUnexpectedInput(
                f"The following input parameters are unexpected: {list(unexpected_types)!r}"
            )

        for i in self.sequences:
            self._nesting_order.update({i.path: i.nesting_order})

        for k, v in self.nesting_order.items():
            if v < 0:
                raise TaskTemplateInvalidNesting(
                    f"`nesting_order` must be >=0 for all keys, but for key {k!r}, value "
                    f"of {v!r} was specified."
                )

        self._defined_input_types = set(input_types)

    def _get_name(self):
        out = f"{self.objective.name}"
        for idx, schema_i in enumerate(self.schemas, start=1):
            need_and = idx < len(self.schemas) and (
                self.schemas[idx].method or self.schemas[idx].implementation
            )
            out += (
                f"{f'_{schema_i.method}' if schema_i.method else ''}"
                f"{f'_{schema_i.implementation}' if schema_i.implementation else ''}"
                f"{f'_and' if need_and else ''}"
            )
        return out

    @staticmethod
    def get_task_unique_names(tasks: List[Task]):
        """Get the unique name of each in a list of tasks.

        Returns
        -------
        list of str

        """

        task_name_rep_idx = get_item_repeat_index(
            tasks,
            item_callable=lambda x: x.name,
            distinguish_singular=True,
        )

        names = []
        for idx, task in enumerate(tasks):
            add_rep = f"_{task_name_rep_idx[idx]}" if task_name_rep_idx[idx] > 0 else ""
            names.append(f"{task.name}{add_rep}")

        return names

    def _get_nesting_order(self, seq):
        """Find the nesting order for a task sequence."""
        return self.nesting_order[seq._get_param_path()] if len(seq.values) > 1 else -1

    def make_persistent(self, workflow):
        """Add all task input data to a persistent workflow and return a record of the
        Zarr parameter group indices for each bit of task data."""

        input_data_indices = {}

        # input_data_indices.update(self.resources.make_persistent(workflow))

        for inp_i in self.inputs:
            input_data_indices.update(inp_i.make_persistent(workflow))

        for seq_i in self.sequences:
            input_data_indices.update(seq_i.make_persistent(workflow))

        return input_data_indices

    def _prepare_persistent_outputs(self, workflow, num_elements):
        # TODO: check that schema is present when adding task? (should this be here?)
        output_data_indices = {}
        for schema in self.schemas:
            for output in schema.outputs:
                output_data_indices[output.typ] = []
                for i in range(num_elements):
                    group_idx = workflow._add_parameter_group(data=None, is_set=False)
                    output_data_indices[output.typ].append(group_idx)

        return output_data_indices

    def prepare_element_resolution(self):

        multiplicities = [
            # {
            #     "multiplicity": 1,
            #     "nesting_order": -1,
            #     "address": self.resources._get_param_path(),
            # }
        ]

        for inp_i in self.inputs:
            multiplicities.append(
                {
                    "multiplicity": 1,
                    "nesting_order": -1,
                    "address": inp_i._get_param_path(),
                }
            )

        for seq_i in self.sequences:
            multiplicities.append(
                {
                    "multiplicity": len(seq_i.values),
                    "nesting_order": self._get_nesting_order(seq_i),
                    "address": seq_i._get_param_path(),
                }
            )

        return multiplicities

    @property
    def index(self):
        if self.workflow_template:
            return self.workflow_template.tasks.index(self)
        else:
            return None

    @classmethod
    def from_spec(cls, spec, all_schemas, all_parameters):
        key = (
            spec.pop("objective"),
            spec.pop("method", None),
            spec.pop("implementation", None),
        )  # TODO: maybe don't mutate spec?
        spec["schemas"] = all_schemas[
            key
        ]  # TODO parse multiple methods/impl not multiple schemas

        sequences = spec.pop("sequences", [])
        inputs_spec = spec.pop("inputs", [])
        input_sources_spec = spec.pop("input_sources", {})
        perturbs_spec = spec.pop("perturbations", {})
        nesting_order = spec.pop("nesting_order", {})

        for k in list(nesting_order.keys()):
            new_k = tuple(k.split("."))
            nesting_order[new_k] = nesting_order.pop(k)

        print(f"nesting_order: {nesting_order}")

        inputs = []
        if isinstance(inputs_spec, dict):
            # transform to a list:
            for input_path, input_val in inputs_spec.items():
                is_sequence = input_path.endswith("[]")
                if is_sequence:
                    input_path = input_path.split("[]")[0]
                input_path = input_path.split(".")
                inputs.append(
                    {
                        "parameter": input_path[0],
                        "path": input_path[1:],
                        "value": input_val[0] if is_sequence else input_val,
                    }
                )
                if is_sequence:
                    seq_path = ["inputs"] + input_path
                    sequences.append(
                        {
                            "path": seq_path,
                            "values": input_val,
                            "nesting_order": nesting_order.get(tuple(seq_path)),
                        }
                    )
        else:
            inputs = inputs_spec

        # add any nesting orders that are defined in the sequences:
        for i in sequences:
            if "nesting_order" in i:
                nesting_order.update({tuple(i["path"]): i["nesting_order"]})

        perturbs = [{"name": p_name, **p} for p_name, p in perturbs_spec.items()]

        spec.update(
            {
                "inputs": [InputValue.from_spec(i, all_parameters) for i in inputs],
                "sequences": [ValueSequence.from_spec(i) for i in sequences],
                "perturbations": [ValuePerturbation.from_spec(i) for i in perturbs],
                "nesting_order": nesting_order,
                "input_sources": {
                    i: [InputSource.from_spec(j) for j in sources]
                    for i, sources in input_sources_spec.items()
                },
            }
        )

        return cls(**spec)

    def get_available_task_input_sources(
        self, source_tasks: Optional[List[Task]] = None
    ) -> List[InputSource]:
        """For each input parameter of this task, generate a list of possible input sources
        that derive from inputs or outputs of this and other provided tasks.

        Note this only produces a subset of available input sources for each input
        parameter; other available input sources may exist from workflow imports."""

        available = {}
        for schema_input in self.all_schema_inputs:
            available[schema_input.typ] = []

            print(f"schema_input: {schema_input.parameter.typ}")

            for src_task_idx, src_task_i in enumerate(source_tasks or []):

                for param_i in src_task_i.provides_parameters:

                    if param_i.typ == schema_input.typ:

                        available[schema_input.typ].append(
                            self.app.InputSource(
                                source_type=self.app.InputSourceType.TASK,
                                task_ref=src_task_i.insert_ID,
                                task_source_type={
                                    "SchemaInput": self.app.TaskSourceType.INPUT,
                                    "SchemaOutput": self.app.TaskSourceType.OUTPUT,
                                }[
                                    param_i.__class__.__name__
                                ],  # TODO: make nicer
                            )
                        )

            if schema_input.typ in self.defined_input_types:
                available[schema_input.typ].append(self.app.InputSource.local())

            if schema_input.default_value is not None:
                available[schema_input.typ].append(self.app.InputSource.default())

        return available

    @property
    def schemas(self):
        return self._schemas

    @property
    def repeats(self):
        return self._repeats

    @property
    def resources(self):
        return self._resources

    @property
    def inputs(self):
        return self._inputs

    @property
    def input_files(self):
        return self._input_files

    @property
    def input_file_generator_sources(self):
        return self._input_file_generator_sources

    @property
    def output_file_parser_sources(self):
        return self._output_file_parser_sources

    @property
    def perturbations(self):
        return self._perturbations

    @property
    def sequences(self):
        return self._sequences

    @property
    def input_sources(self):
        return self._input_sources

    @property
    def input_source_mode(self):
        return self._input_source_mode

    @property
    def insert_ID(self):
        return self._insert_ID

    @property
    def dir_name(self):
        "Artefact directory name."
        return self._dir_name

    @property
    def nesting_order(self):
        return self._nesting_order

    @property
    def groups(self):
        return self._groups

    @property
    def name(self):
        return self._name

    @property
    def objective(self):
        return self.schemas[0].objective

    @property
    def all_schema_inputs(self) -> Tuple[SchemaInput]:
        return tuple(inp_j for schema_i in self.schemas for inp_j in schema_i.inputs)

    @property
    def all_schema_outputs(self) -> Tuple[SchemaOutput]:
        return tuple(inp_j for schema_i in self.schemas for inp_j in schema_i.outputs)

    @property
    def all_schema_input_types(self):
        """Get the set of all schema input types (over all specified schemas)."""
        return {inp_j for schema_i in self.schemas for inp_j in schema_i.input_types}

    @property
    def all_schema_output_types(self):
        """Get the set of all schema output types (over all specified schemas)."""
        return {out_j for schema_i in self.schemas for out_j in schema_i.output_types}

    @property
    def universal_input_types(self):
        """Get input types that are associated with all schemas"""

    @property
    def non_universal_input_types(self):
        """Get input types for each schema that are non-universal."""

    @property
    def defined_input_types(self):
        return self._defined_input_types

    @property
    def undefined_input_types(self):
        return self.all_schema_input_types - self.defined_input_types

    @property
    def undefined_inputs(self):
        return [
            inp_j
            for schema_i in self.schemas
            for inp_j in schema_i.inputs
            if inp_j.typ in self.undefined_input_types
        ]

    @property
    def unsourced_inputs(self):
        """Get schema input types for which no input sources are currently specified."""
        return self.all_schema_input_types - set(self.input_sources.keys())

    @property
    def provides_parameters(self):
        return tuple(j for schema in self.schemas for j in schema.provides_parameters)

    def get_sub_parameter_input_values(self):
        return [i for i in self.inputs if i.is_sub_value]

    def get_non_sub_parameter_input_values(self):
        return [i for i in self.inputs if not i.is_sub_value]

    def add_group(
        self, name: str, where: ElementFilter, group_by_distinct: ParameterPath
    ):
        group = ElementGroup(name=name, where=where, group_by_distinct=group_by_distinct)
        self.groups.add_object(group)

    def get_input_multiplicities(self, missing_multiplicities=None):
        """Get multiplicities for all inputs."""

        if self.undefined_input_types:
            missing_inputs = self.undefined_input_types - set(
                (missing_multiplicities or {}).keys()
            )
            if missing_inputs:
                raise MissingInputs(
                    f"The following inputs are not assigned values, so task input "
                    f"multiplicities cannot be resolved: {list(missing_inputs)!r}."
                )

        input_multiplicities = []
        for i in self.input_values:
            if i.sequences:
                for seq in i.sequences:
                    address = tuple([i.parameter.typ] + seq.address)
                    address_str = ".".join(address)
                    input_multiplicities.append(
                        {
                            "address": address,
                            "multiplicity": len(seq.values),
                            "nesting_order": self.nesting_order[address_str],
                        }
                    )
            else:
                input_multiplicities.append(
                    {
                        "address": (i.parameter.typ,),
                        "multiplicity": 1,
                        "nesting_order": -1,
                    }
                )

        for inp_type, multi in (missing_multiplicities or {}).items():
            input_multiplicities.append(
                {
                    "address": (inp_type,),
                    "multiplicity": multi,
                    "nesting_order": self.nesting_order[inp_type],
                }
            )

        return input_multiplicities


class WorkflowTask:
    """Class to represent a Task that is bound to a Workflow."""

    def __init__(
        self,
        template: Task,
        element_indices: List,
        index: int,
        workflow: Workflow,
    ):

        self._template = template
        self._element_indices = element_indices
        self._workflow = workflow
        self._index = index

    @property
    def template(self):
        return self._template

    @property
    def element_indices(self):
        return self._element_indices

    @property
    def elements(self):
        return [self.workflow.elements[i] for i in self.element_indices]

    @property
    def workflow(self):
        return self._workflow

    @property
    def num_elements(self):
        return len(self.element_indices)

    @property
    def index(self):
        """Zero-based position within the workflow. Uses initial index if appending to the
        workflow is not complete."""
        return self._index

    @property
    def name(self):
        return self.template.name

    @property
    def insert_ID(self):
        return self.template.insert_ID

    @property
    def dir_name(self):
        return self.template.dir_name

    @property
    def unique_name(self):
        return self.workflow.get_task_unique_names()[self.index]



from __future__ import annotations
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Union
from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike


from hpcflow.sdk.core.environment import Environment
from hpcflow.sdk.core.utils import search_dir_files_by_regex


@dataclass
class FileSpec(JSONLike):

    _validation_schema = "files_spec_schema.yaml"
    _child_objects = (ChildObjectSpec(name="name", class_name="FileNameSpec"),)

    label: str
    name: str
    _hash_value: Optional[str] = field(default=None, repr=False)

    def __post_init__(self):
        self.name = FileNameSpec(self.name) if isinstance(self.name, str) else self.name

    def value(self, directory=None):
        return self.name.value(directory)

    @property
    def stem(self):
        return self.name.stem

    @property
    def ext(self):
        return self.name.ext


class FileNameSpec(JSONLike):
    def __init__(self, name, args=None, is_regex=False):
        self.name = name
        self.args = args
        self.is_regex = is_regex

    @property
    def stem(self):
        return FileNameStem(self)

    @property
    def ext(self):
        return FileNameExt(self)

    def value(self, directory=None):
        format_args = [i.value(directory) for i in self.args or []]
        value = self.name.format(*format_args)
        if self.is_regex:
            value = search_dir_files_by_regex(value, group=0, directory=directory)
        return value

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"


@dataclass
class FileNameStem(JSONLike):
    file_name: FileNameSpec

    def value(self, directory=None):
        return Path(self.file_name.value(directory)).stem


@dataclass
class FileNameExt(JSONLike):
    file_name: FileNameSpec

    def value(self, directory=None):
        return Path(self.file_name.value(directory)).suffix


@dataclass
class InputFileGenerator(JSONLike):

    _child_objects = (
        ChildObjectSpec(
            name="input_file",
            class_name="FileSpec",
            shared_data_primary_key="label",
            shared_data_name="command_files",
        ),
        ChildObjectSpec(
            name="inputs",
            class_name="Parameter",
            is_multiple=True,
            json_like_name="from_inputs",
            shared_data_primary_key="typ",
            shared_data_name="parameters",
        ),
    )

    input_file: FileSpec
    inputs: List[Parameter]
    environment: Environment = None

    # @classmethod
    # def from_spec(cls, label, info, parameters, cmd_files):
    #     input_file = [i for i in cmd_files if i.label == label][0]
    #     inputs = [parameters[typ] for typ in info["from_inputs"]]
    #     return cls(input_file, inputs)


@dataclass
class OutputFileParser(JSONLike):

    _child_objects = (
        ChildObjectSpec(
            name="output",
            class_name="Parameter",
            shared_data_name="parameters",
            shared_data_primary_key="typ",
        ),
        ChildObjectSpec(
            name="output_files",
            json_like_name="from_files",
            class_name="FileSpec",
            is_multiple=True,
            shared_data_primary_key="label",
            shared_data_name="command_files",
        ),
    )

    output: Parameter
    output_files: List[FileSpec]
    environment: Environment = None
    inputs: List[str] = None
    options: Dict = None

    # @classmethod
    # def from_spec(cls, param_typ, info, parameters, cmd_files):
    #     output = parameters[param_typ]
    #     output_files = [
    #         [j for j in cmd_files if j.label == label][0]
    #         for label in info["from_files"]
    #     ]
    #     return cls(output, output_files)


class _FileContentsSpecifier(JSONLike):
    """Class to represent the contents of a file, either via a file-system path or directly."""

    def __init__(
        self, path: Union[Path, str] = None, contents: str = None, extension: str = ""
    ):
        self.path = Path(path) if path is not None else path
        self._contents = contents
        self.extension = extension

        if (path is not None and contents is not None) or (
            path is None and contents is None
        ):
            raise ValueError("Specify exactly one of `path` and `contents`.")

    @property
    def contents(self):
        if self.path is not None:
            with self.path.open("r") as fh:
                contents = fh.read()
        else:
            contents = self._contents
        return contents


class InputFile(_FileContentsSpecifier):
    def __init__(
        self,
        file: Union[FileSpec, str],
        path: Union[Path, str] = None,
        contents: str = None,
        extension: str = "",
    ):
        super().__init__(path, contents, extension)
        self.file = file
        if not isinstance(self.file, FileSpec):
            self.file = self.app.command_files[self.file]


class InputFileGeneratorSource(_FileContentsSpecifier):
    def __init__(
        self,
        generator: InputFileGenerator,
        path: Union[Path, str] = None,
        contents: str = None,
        extension: str = "",
    ):
        super().__init__(path, contents, extension)
        self.generator = generator


class OutputFileParserSource(_FileContentsSpecifier):
    def __init__(
        self,
        parser: OutputFileParser,
        path: Union[Path, str] = None,
        contents: str = None,
        extension: str = "",
    ):
        super().__init__(path, contents, extension)
        self.parser = parser



import pytest

from hpcflow.api import Action, Command, TaskObjective, TaskSchema
from hpcflow.sdk.core.errors import InvalidIdentifier


@pytest.fixture
def action_a1():
    return Action(commands=[Command("ls")], environments=[])


@pytest.fixture
def schema_s1_kwargs(action_a1):
    return {"objective": TaskObjective("t1"), "actions": [action_a1]}


def test_task_schema_equality():
    t1a = TaskSchema("t1", actions=[])
    t1b = TaskSchema("t1", actions=[])
    assert t1a == t1b


def test_init_with_str_objective(action_a1):
    obj_str = "t1"
    obj = TaskObjective(obj_str)
    common = {"actions": [action_a1]}
    assert TaskSchema(obj_str, **common) == TaskSchema(obj, **common)


def test_init_with_method_with_underscore(schema_s1_kwargs):
    TaskSchema(method="my_method", **schema_s1_kwargs)


def test_raise_on_invalid_method_digit(schema_s1_kwargs):
    with pytest.raises(InvalidIdentifier):
        TaskSchema(method="9", **schema_s1_kwargs)


def test_raise_on_invalid_method_space(schema_s1_kwargs):
    with pytest.raises(InvalidIdentifier):
        TaskSchema(method="my method", **schema_s1_kwargs)


def test_raise_on_invalid_method_non_alpha_numeric(schema_s1_kwargs):
    with pytest.raises(InvalidIdentifier):
        TaskSchema(method="_mymethod", **schema_s1_kwargs)


def test_method_lowercasing(schema_s1_kwargs):
    assert TaskSchema(method="MyMethod", **schema_s1_kwargs) == TaskSchema(
        method="mymethod", **schema_s1_kwargs
    )



from __future__ import annotations
from dataclasses import dataclass
from typing import Dict, List, Optional

from valida.conditions import ConditionLike
from hpcflow.sdk.core.zarr_io import zarr_decode
from hpcflow.sdk.core.utils import (
    check_valid_py_identifier,
    get_in_container,
    get_relative_path,
    set_in_container,
)


@dataclass
class ElementInputs:

    element: Element

    def __repr__(self):
        return f"{self.__class__.__name__}(" f"{', '.join(self._get_input_names())}" f")"

    def _get_input_names(self):
        return sorted(self.element.task.template.all_schema_input_types)

    def __getattr__(self, name):
        if name not in self._get_input_names():
            raise ValueError(f"No input named {name!r}.")
        return self.element.get(("inputs", name))

    def __dir__(self):
        return super().__dir__() + self._get_input_names()


@dataclass
class ElementOutputs:

    element: Element

    def _get_output_names(self):
        return list(self.element.task.template.all_schema_output_types)

    def __getattr__(self, name):
        if name not in self._get_output_names():
            raise ValueError(f"No output named {name!r}.")
        return self.element.get(("outputs", name))

    def __dir__(self):
        return super().__dir__() + self._get_output_names()


@dataclass
class Element:
    task: Task
    data_index: Dict

    @property
    def workflow(self):
        return self.task.workflow

    @property
    def inputs(self):
        return ElementInputs(self)

    @property
    def outputs(self):
        return ElementOutputs(self)

    def __post_init__(self):
        # ensure sorted from smallest to largest path:
        self.data_index = dict(
            sorted(
                {tuple(k.split(".")): v for k, v in self.data_index.items()}.items(),
                key=lambda x: len(x[0]),
            )
        )

    def _path_to_parameter(self, path):
        if len(path) != 2:
            return

        if path[0] == "inputs":
            for i in self.task.template.schemas:
                for j in i.inputs:
                    if j.parameter.typ == path[1]:
                        return j.parameter

        elif path[0] == "outputs":
            for i in self.task.template.schemas:
                for j in i.outputs:
                    if j.parameter.typ == path[1]:
                        return j.parameter

    def get(self, path=None):
        """Get element data from the persistent store."""

        path = path or []
        parameter = self._path_to_parameter(path)
        print(f"parameter: {parameter}")
        current_value = None
        for path_i, data_idx_i in self.data_index.items():

            is_parent = False
            is_update = False
            try:
                rel_path = get_relative_path(path, path_i)
                is_parent = True
            except ValueError:
                try:
                    update_path = get_relative_path(path_i, path)
                    is_update = True

                except ValueError:
                    continue

            zarr_group = self.workflow.get_zarr_parameter_group(data_idx_i)
            data = zarr_decode(zarr_group)

            if is_parent:
                # replace current value:
                try:
                    current_value = get_in_container(data, rel_path)
                except (KeyError, IndexError):
                    continue

            elif is_update:
                # update sub-part of current value
                current_value = current_value or {}
                set_in_container(current_value, update_path, data, ensure_path=True)

        if parameter and parameter._value_class:
            current_value = parameter._value_class(**current_value)

        return current_value


@dataclass
class ElementFilter:

    parameter_path: ParameterPath
    condition: ConditionLike


@dataclass
class ElementGroup:

    name: str
    where: Optional[ElementFilter] = None
    group_by_distinct: Optional[ParameterPath] = None

    def __post_init__(self):
        self.name = check_valid_py_identifier(self.name)


@dataclass
class ElementRepeats:

    number: int
    where: Optional[ElementFilter] = None



import pytest

from hpcflow.api import (
    ActionScope,
    Parameter,
    InputValue,
    Task,
    TaskSchema,
    ValueSequence,
    ResourceSpec,
)
from hpcflow.sdk.core.errors import InputValueDuplicateSequenceAddress


@pytest.fixture
def param_p1():
    return Parameter("p1")


def test_fix_trailing_path_delimiter(param_p1):
    iv1 = InputValue(parameter=param_p1, value=101, path="a.")
    iv2 = InputValue(parameter=param_p1, value=101, path="a")
    assert iv1.path == iv2.path


def test_fix_single_path_delimiter(param_p1):
    iv1 = InputValue(parameter=param_p1, value=101, path=".")
    iv2 = InputValue(parameter=param_p1, value=101)
    assert iv1.path == iv2.path


def test_get_param_path_without_path(param_p1):
    iv1 = InputValue(parameter=param_p1, value=101)
    assert iv1._get_param_path() == "inputs.p1"


def test_get_param_path_with_single_element_path(param_p1):
    iv1 = InputValue(parameter=param_p1, value=101, path="a")
    assert iv1._get_param_path() == "inputs.p1.a"


def test_get_param_path_with_multi_element_path(param_p1):
    iv1 = InputValue(parameter=param_p1, value=101, path="a.b")
    assert iv1._get_param_path() == "inputs.p1.a.b"


def test_get_param_path_with_empty_path(param_p1):
    iv1 = InputValue(parameter=param_p1, value=101, path="")
    assert iv1._get_param_path() == "inputs.p1"


def test_resource_spec_get_param_path():
    rs1 = ResourceSpec()
    assert rs1._get_param_path() == "resources"


def test_resource_spec_get_param_path_scope_any_with_single_kwarg():
    rs1 = ResourceSpec(scratch="local")
    assert rs1._get_param_path() == "resources"


def test_resources_spec_get_param_path_scope_main():
    rs1 = ResourceSpec(scope=ActionScope.main())
    assert rs1._get_param_path() == "resources.main"


def test_resources_spec_get_param_path_scope_with_kwargs():
    rs1 = ResourceSpec(scope=ActionScope.input_file_generator(file="file1"))
    assert rs1._get_param_path() == "resources.input_file_generator[file=file1]"


def test_resources_spec_get_param_path_scope_with_no_kwargs():
    rs1 = ResourceSpec(scope=ActionScope.input_file_generator())
    assert rs1._get_param_path() == "resources.input_file_generator"


# def test_raise_on_duplicate_input_value_sequence_address(param_p1):
#     with pytest.raises(InputValueDuplicateSequenceAddress):
#         InputValue(
#             parameter=param_p1,
#             value={"A": 1},
#             sequences=[
#                 ValueSequence(values=[1, 2, 3], path=("A",), nesting_order=0),
#                 ValueSequence(values=[4, 5, 6], path=("A",), nesting_order=0),
#             ],
#         )

#     s1 = TaskSchema("t1", inputs=[param_p1], actions=[])
#     t1 = Task(schemas=[s1], inputs=[InputValue(param_p1, value=101)])


# def test_raise_on_duplicate_input_value_sequence_address_empty():
#     p1 = Parameter("p1")
#     with pytest.raises(InputValueDuplicateSequenceAddress):
#         InputValue(
#             parameter=p1,
#             sequences=[ValueSequence(values=[1, 2, 3]), ValueSequence(values=[4, 5, 6])],
#         )



from __future__ import annotations
from dataclasses import dataclass
import enum
import re
from typing import List, Optional, Tuple

from valida.conditions import ConditionLike

from hpcflow.sdk.core.command_files import InputFileGenerator, OutputFileParser
from hpcflow.sdk.core.commands import Command
from hpcflow.sdk.core.environment import Environment
from hpcflow.sdk.core.errors import MissingCompatibleActionEnvironment
from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike


ACTION_SCOPE_REGEX = r"(\w*)(?:\[(.*)\])?"


class ActionScopeType(enum.Enum):

    ANY = 0
    MAIN = 1
    PROCESSING = 2
    INPUT_FILE_GENERATOR = 3
    OUTPUT_FILE_PARSER = 4


ACTION_SCOPE_ALLOWED_KWARGS = {
    ActionScopeType.ANY.name: set(),
    ActionScopeType.MAIN.name: set(),
    ActionScopeType.PROCESSING.name: set(),
    ActionScopeType.INPUT_FILE_GENERATOR.name: {"file"},
    ActionScopeType.OUTPUT_FILE_PARSER.name: {"output"},
}


class ActionScope(JSONLike):
    """Class to represent the identification of a subset of task schema actions by a
    filtering process.
    """

    _child_objects = (
        ChildObjectSpec(
            name="typ",
            json_like_name="type",
            class_name="ActionScopeType",
            is_enum=True,
        ),
    )

    def __init__(self, typ: ActionScopeType, **kwargs):
        self.typ = typ
        self.kwargs = {k: v for k, v in kwargs.items() if v is not None}

        bad_keys = set(kwargs.keys()) - ACTION_SCOPE_ALLOWED_KWARGS[self.typ.name]
        if bad_keys:
            raise TypeError(
                f"The following keyword arguments are unknown for ActionScopeType "
                f"{self.typ.name}: {bad_keys}."
            )

    def __repr__(self):
        kwargs_str = ""
        if self.kwargs:
            kwargs_str = ", ".join(f"{k}={v!r}" for k, v in self.kwargs.items())
        return f"{self.__class__.__name__}.{self.typ.name.lower()}({kwargs_str})"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.typ is other.typ and self.kwargs == other.kwargs:
            return True
        return False

    @classmethod
    def _parse_from_string(cls, string):
        typ_str, kwargs_str = re.search(ACTION_SCOPE_REGEX, string).groups()
        kwargs = {}
        if kwargs_str:
            for i in kwargs_str.split(","):
                name, val = i.split("=")
                kwargs[name.strip()] = val.strip()
        return {"type": typ_str, **kwargs}

    def to_string(self):
        kwargs_str = ""
        if self.kwargs:
            kwargs_str = "[" + ", ".join(f"{k}={v}" for k, v in self.kwargs.items()) + "]"
        return f"{self.typ.name.lower()}{kwargs_str}"

    @classmethod
    def from_json_like(cls, json_like, shared_data=None):
        if isinstance(json_like, str):
            json_like = cls._parse_from_string(json_like)
        else:
            typ = json_like.pop("type")
            json_like = {"type": typ, **json_like.pop("kwargs", {})}
        return super().from_json_like(json_like, shared_data)

    @classmethod
    def any(cls):
        return cls(typ=ActionScopeType.ANY)

    @classmethod
    def main(cls):
        return cls(typ=ActionScopeType.MAIN)

    @classmethod
    def processing(cls):
        return cls(typ=ActionScopeType.PROCESSING)

    @classmethod
    def input_file_generator(cls, file=None):
        return cls(typ=ActionScopeType.INPUT_FILE_GENERATOR, file=file)

    @classmethod
    def output_file_parser(cls, output=None):
        return cls(typ=ActionScopeType.OUTPUT_FILE_PARSER, output=output)


@dataclass
class ActionEnvironment(JSONLike):

    _child_objects = (
        ChildObjectSpec(
            name="scope",
            class_name="ActionScope",
        ),
        ChildObjectSpec(
            name="environment",
            class_name="Environment",
            shared_data_name="envs",
            shared_data_primary_key="name",
        ),
    )

    environment: Environment
    scope: ActionScope

    @classmethod
    def prepare_from_json_like(cls, json_like):

        print(f"ActEnv.prep: json_like {json_like}")

        json_like = {
            "scope": {
                "typ": json_like["scope"],
                "kwargs": {
                    k: v
                    for k, v in json_like.items()
                    if k not in ("environment", "scope")
                },
            },
            "environment": json_like["environment"],
        }
        return super().prepare_from_json_like(json_like)


@dataclass
class ActionCondition:
    """Class to represent a condition that must be met if an action is to be included."""

    path: List[str]
    condition: Optional[ConditionLike] = None


class Action(JSONLike):
    """"""

    _child_objects = (
        ChildObjectSpec(
            name="commands",
            class_name="Command",
            is_multiple=True,
        ),
        ChildObjectSpec(
            name="input_file_generators",
            json_like_name="input_files",
            is_multiple=True,
            class_name="InputFileGenerator",
            dict_key_attr="input_file",
        ),
        ChildObjectSpec(
            name="output_file_parsers",
            json_like_name="outputs",
            is_multiple=True,
            class_name="OutputFileParser",
            dict_key_attr="output",
        ),
        ChildObjectSpec(
            name="environments",
            class_name="ActionEnvironment",
            is_multiple=True,
        ),
    )

    def __init__(
        self,
        commands: List[Command],
        environments: List[ActionEnvironment],
        input_file_generators: Optional[List[InputFileGenerator]] = None,
        output_file_parsers: Optional[List[OutputFileParser]] = None,
        conditions: Optional[List[ActionCondition]] = None,
    ):
        self.commands = commands
        self.environments = environments
        self.input_file_generators = input_file_generators or []
        self.output_file_parsers = output_file_parsers or []
        self.conditions = conditions or []

    def __eq__(self, other):
        if type(other) is not self.__class__:
            return False
        if (
            self.commands == other.commands
            and self.environments == other.environments
            and self.input_file_generators == other.input_file_generators
            and self.output_file_parsers == other.output_file_parsers
            and self.conditions == other.conditions
        ):
            return True
        return False

    def get_parameter_dependence(self, parameter: SchemaParameter):
        """Find if/where a given parameter is used by the action."""
        writer_files = [
            i.input_file
            for i in self.input_file_generators
            if parameter.parameter in i.inputs
        ]  # names of input files whose generation requires this parameter
        commands = []  # TODO: indices of commands in which this parameter appears
        out = {"input_file_writers": writer_files, "commands": commands}
        return out

    def get_resolved_action_env(
        self,
        relevant_scopes: Tuple[ActionScopeType],
        input_file_generator: InputFileGenerator = None,
        output_file_parser: OutputFileParser = None,
        commands: List[Command] = None,
    ):
        print(f"relevant_scopes: {relevant_scopes}")
        print(f"self.environments: {self.environments}")
        print(
            f"[i.scope.typ for i in self.environments]: {[i.scope.typ for i in self.environments]}"
        )
        possible = [
            i.scope.type for i in self.environments if i.scope.typ in relevant_scopes
        ]
        print(f"possible: {possible}")
        if not possible:
            if input_file_generator:
                msg = f"input file generator {input_file_generator!r}."
            elif output_file_parser:
                msg = f"output file parser {output_file_parser!r}"
            else:
                msg = f"commands {commands!r}"
            raise MissingCompatibleActionEnvironment(
                f"No compatible environment is specified for the {msg}."
            )

        # sort by scope specificity:
        possible_srt = sorted(possible, key=lambda i: i.scope.value, reverse=True)
        return possible_srt[0]

    def get_input_file_generator_action_env(
        self, input_file_generator: InputFileGenerator
    ):
        return self.get_resolved_action_env(
            relevant_scopes=(
                ActionScopeType.ANY,
                ActionScopeType.PROCESSING,
                ActionScopeType.INPUT_FILE_GENERATOR,
            ),
            input_file_generator=input_file_generator,
        )

    def get_output_file_parser_action_env(self, output_file_parser: OutputFileParser):
        return self.get_resolved_action_env(
            relevant_scopes=(
                ActionScopeType.ANY,
                ActionScopeType.PROCESSING,
                ActionScopeType.OUTPUT_FILE_PARSER,
            ),
            output_file_parser=output_file_parser,
        )

    def get_commands_action_env(self):
        return self.get_resolved_action_env(
            relevant_scopes=(ActionScopeType.ANY, ActionScopeType.MAIN),
            commands=self.commands,
        )

    def resolve_actions(self):

        cmd_acts = []
        for i in self.input_file_generators:
            act_i = InputFileGeneratorAction(
                input_file_generator=i,
                conditions=self.conditions,
                environment=self.get_input_file_generator_action_env(i),
            )
            cmd_acts.append(act_i)

        cmd_acts.append(
            CommandsAction(
                commands=self.commands,
                environment=self.get_commands_action_env(),
                conditions=self.conditions,
            )
        )

        for i in self.output_file_parsers:
            act_i = OutputFileParserAction(
                output_file_parser=i,
                environment=self.get_output_file_parser_action_env(i),
                conditions=self.conditions,
            )
            cmd_acts.append(act_i)

        return cmd_acts


@dataclass
class ResolvedAction:
    """"""

    environment: Environment
    conditions: List[ActionCondition]

    def __post_init__(self):
        # select correct environment
        pass


@dataclass
class CommandsAction(ResolvedAction):
    """Represents an action without any associated input file generators and output
    parsers."""

    commands: List[Command]


@dataclass
class InputFileGeneratorAction(ResolvedAction):
    input_file_generator: InputFileGenerator

    def __post_init__(self):
        self.conditions = (
            []
        )  # TODO: add a condition, according to non-presence of input file?


@dataclass
class OutputFileParserAction(ResolvedAction):
    output_file_parser: OutputFileParser



# import pytest

# from hpcflow.utils import (
#     get_duplicate_items,
#     check_valid_py_identifier,
#     group_by_dict_key_values,
# )


# def test_get_list_duplicate_items_no_duplicates():
#     lst = [1, 2, 3]
#     assert not get_duplicate_items(lst)


# def test_get_list_duplicate_items_one_duplicate():
#     lst = [1, 1, 3]
#     assert get_duplicate_items(lst) == [1]


# def test_get_list_duplicate_items_all_duplicates():
#     lst = [1, 1, 1]
#     assert get_duplicate_items(lst) == [1]


# def test_raise_check_valid_py_identifier_empty_str():
#     with pytest.raises(ValueError):
#         check_valid_py_identifier("")


# def test_raise_check_valid_py_identifier_start_digit():
#     with pytest.raises(ValueError):
#         check_valid_py_identifier("9sdj")


# def test_raise_check_valid_py_identifier_single_digit():
#     with pytest.raises(ValueError):
#         check_valid_py_identifier("9")


# def test_raise_check_valid_py_identifier_py_keyword():
#     with pytest.raises(ValueError):
#         check_valid_py_identifier("if")


# def test_expected_return_check_valid_py_identifier_all_latin_alpha():
#     assert check_valid_py_identifier("abc") == "abc"


# def test_expected_return_check_valid_py_identifier_all_latin_alphanumeric():
#     assert check_valid_py_identifier("abc123") == "abc123"


# def test_expected_return_check_valid_py_identifier_all_greek_alpha():
#     assert check_valid_py_identifier("αβγ") == "αβγ"


# def test_check_valid_py_identifier_case_insensitivity():
#     assert (
#         check_valid_py_identifier("abc012")
#         == check_valid_py_identifier("ABC012")
#         == check_valid_py_identifier("aBc012")
#         == "abc012"
#     )


# def test_expected_return_group_by_dict_key_values_single_key_items_single_key_passed():
#     item_1 = {"b": 1}
#     item_2 = {"b": 2}
#     item_3 = {"b": 1}
#     assert group_by_dict_key_values([item_1, item_2, item_3], "b") == [
#         [item_1, item_3],
#         [item_2],
#     ]


# def test_expected_return_group_by_dict_key_values_multi_key_items_single_key_passed():
#     item_1 = {"a": 9, "b": 1}
#     item_2 = {"a": 8, "b": 2}
#     item_3 = {"a": 9, "b": 1}
#     assert group_by_dict_key_values([item_1, item_2, item_3], "b") == [
#         [item_1, item_3],
#         [item_2],
#     ]


# def test_expected_return_group_by_dict_key_values_multi_key_items_multi_key_passed_two_groups():
#     item_1 = {"a": 9, "b": 1}
#     item_2 = {"a": 8, "b": 2}
#     item_3 = {"a": 9, "b": 1}
#     assert group_by_dict_key_values([item_1, item_2, item_3], "a", "b") == [
#         [item_1, item_3],
#         [item_2],
#     ]


# def test_expected_return_group_by_dict_key_values_multi_key_items_multi_key_passed_three_groups():
#     item_1 = {"a": 9, "b": 1}
#     item_2 = {"a": 9, "b": 2}
#     item_3 = {"a": 8, "b": 1}
#     assert group_by_dict_key_values([item_1, item_2, item_3], "a", "b") == [
#         [item_1],
#         [item_2],
#         [item_3],
#     ]


# def test_expected_return_group_by_dict_key_values_multi_key_items_multi_key_passed_one_group():
#     item_1 = {"a": 9, "b": 1}
#     item_2 = {"a": 9, "b": 1}
#     item_3 = {"a": 9, "b": 1}
#     assert group_by_dict_key_values([item_1, item_2, item_3], "a", "b") == [
#         [item_1, item_2, item_3]
#     ]


# def test_expected_return_group_by_dict_key_values_excluded_items_for_missing_keys_first_item():
#     item_1 = {"a": 9}
#     item_2 = {"a": 9, "b": 1}
#     item_3 = {"a": 9, "b": 1}
#     assert group_by_dict_key_values([item_1, item_2, item_3], "a", "b") == [
#         [item_1],
#         [item_2, item_3],
#     ]


# def test_expected_return_group_by_dict_key_values_excluded_items_for_missing_keys_second_item():
#     item_1 = {"a": 9, "b": 1}
#     item_2 = {"a": 9}
#     item_3 = {"a": 9, "b": 1}
#     assert group_by_dict_key_values([item_1, item_2, item_3], "a", "b") == [
#         [item_1, item_3],
#         [item_2],
#     ]



from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import List, Optional
from pprint import pprint
from warnings import warn

import zarr

from .element import Element
from .json_like import ChildObjectSpec, JSONLike
from .zarr_io import zarr_encode
from .object_list import WorkflowTaskList
from .parameters import InputSource
from .loop import Loop
from .task import Task, WorkflowTask
from .task_schema import TaskSchema
from .utils import group_by_dict_key_values, read_YAML_file
from .errors import InvalidInputSourceTaskReference, WorkflowNotFoundError

TS_FMT = r"%Y.%m.%d_%H:%M:%S_%z"
TS_NAME_FMT = r"%Y-%m-%d_%H%M%S"


# class WorkflowTemplateOld(JSONLike):

#     _child_objects = (
#         ChildObjectSpec(
#             name="tasks",
#             class_name="TaskTemplate",
#             is_multiple=True,
#         ),
#     )

#     def __init__(
#         self,
#         tasks: Optional[List[TaskTemplate]] = None,
#         loops: Optional[List[Loop]] = None,
#     ):

#         self.parameter_data = []
#         self.parameter_mapping = []
#         self.elements = []
#         self.tasks = TaskList([])
#         self.element_indices = []
#         self.name_repeat_indices = []

#         for task in tasks or []:
#             self.add_task(task)

#     @classmethod
#     def from_YAML_file(cls, path):
#         dat = read_YAML_file(path)
#         return cls.from_json_like(dat, shared_data=cls.app.app_data)

#     @classmethod
#     def from_JSON_file(cls, path):
#         dat = read_JSON_file(path)
#         return cls.from_json_like(dat, shared_data=cls.app.app_data)

#     def get_possible_input_sources(
#         self, schema_input: SchemaInput, new_task: TaskTemplate, new_index: int
#     ):
#         """Enumerate the possible sources for an input of a new task, given a proposed
#         placement of that task."""

#         # Get parameters provided by tasks up to `new_index`:
#         task_sources = {}
#         for task in self.tasks[:new_index]:
#             provided = tuple(
#                 i for i in task.template.provides_parameters if i.typ == schema_input.typ
#             )
#             if provided:
#                 task_sources.update({(task.index, task.unique_name): provided})

#         out = {
#             "imports": {},
#             "tasks": task_sources,
#             "has_local": schema_input.typ in new_task.defined_input_types,
#             # TODO: there *might* be local definition of a parameter in the form of the input files/writers specified?
#             "has_default": schema_input.default_value is not None,
#         }
#         return out

#     def ensure_input_sources(self, new_task: TaskTemplate, new_index: int):
#         """Check valid input sources are specified for a new task to be added to the
#         workflow in a given position. If none are specified, set them according to the
#         default behaviour."""

#         all_sources = {}
#         for schema_input in new_task.all_schema_inputs:

#             all_sources.update(
#                 {
#                     schema_input.typ: self.get_possible_input_sources(
#                         schema_input, new_task, new_index
#                     )
#                 }
#             )

#             # if sources are specified, check they are resolvable:
#             for specified_source in new_task.input_sources.get(schema_input.typ, []):
#                 specified_source.validate(schema_input, new_task, self)

#         print(f"\nall_sources: {all_sources}")
#         print(f"new_task.unsourced_inputs: {new_task.unsourced_inputs}")

#         # if an input is not specified at all in the `inputs` dict (what about when list?),
#         # then check if there is an input files entry for associated inputs,
#         # if there is

#         # set source for any unsourced inputs:
#         for input_type in new_task.unsourced_inputs:
#             inp_i_sources = all_sources[input_type]

#             print(f"sources: {inp_i_sources}")

#             # input may not be required

#             # set source for this input:
#             if inp_i_sources["has_local"]:
#                 new_sources = [InputSource("local")]

#             elif inp_i_sources["tasks"]:
#                 # we can only take parameters with implicit propagation mode:
#                 params_info = []
#                 for (task_idx, task_name), params in inp_i_sources["tasks"].items():
#                     for i in params:
#                         if i.propagation_mode == ParameterPropagationMode.IMPLICIT:
#                             params_info.append((i.input_or_output, task_idx, task_name))

#                 # sort by output/input (output first), then by task index (highest first):
#                 params_info = sorted(
#                     params_info, key=itemgetter(0, 1), reverse=True
#                 )  # TODO: TEST THIS!!!
#                 new_sources = [
#                     InputSource(f"tasks.{params_info[0][2]}.{params_info[0][0]}s")
#                 ]

#             else:
#                 # input may not need defining (if all associated input files are passed
#                 # and the input does not appear in any action commands)
#                 new_sources = None

#             new_task.input_sources.update({input_type: new_sources})

#     def add_task(self, task_template: Task):

#         # TODO: can't do this check yet because required inputs of different elements may be different
#         # e.g. if an input file is passed for some elements but not others.

#         self.ensure_input_sources(
#             task_template, len(self.tasks)
#         )  # modifies task_template.input_sources
#         # at this point the source for each input should be decided and well-defined.

#         element_indices = []

#         add_sequences = [  # treat the base inputs and resources as single-item sequences:
#             ValueSequence(
#                 path=["inputs"],
#                 values=[
#                     {
#                         param.parameter.typ: param.value
#                         for param in task_template.get_non_sub_parameter_input_values()
#                     }
#                 ],
#                 nesting_order=-1,
#             ),
#             ValueSequence(
#                 path=["resources"],
#                 values=[task_template.resources],
#                 nesting_order=-1,
#             ),
#         ]

#         # treat sub-parameter input values as single-item sequences
#         for i in task_template.get_sub_parameter_input_values():
#             add_sequences.append(
#                 ValueSequence(
#                     path=["inputs", i.parameter.typ] + i.path,
#                     values=[i.value],
#                     nesting_order=-1,
#                 )
#             )

#         multi = []
#         input_map_indices = {}

#         sequences = add_sequences + task_template.sequences
#         for i in sequences:
#             # add each sequence data:
#             next_param_idx = len(self.parameter_data)
#             num_values = len(i.values)
#             self.parameter_data.extend([{"is_set": True, "data": j} for j in i.values])
#             self.parameter_mapping.append(
#                 list(range(next_param_idx, next_param_idx + num_values))
#             )
#             param_map_idx = len(self.parameter_mapping) - 1
#             input_map_indices[tuple(i.path)] = param_map_idx
#             nesting_order_i = (
#                 task_template.nesting_order[tuple(i.path)] if num_values > 1 else -1
#             )
#             multi.append(
#                 {
#                     "multiplicity": num_values,
#                     "nesting_order": nesting_order_i,
#                     "address": i.path,
#                 }
#             )

#         init_multi = WorkflowTemplateOld.resolve_initial_elements(multi)
#         output_map_indices = {}
#         num_elems = len(init_multi)
#         for schema in task_template.schemas:
#             for output in schema.outputs:
#                 next_dat_idx = len(self.parameter_data)
#                 next_map_idx = len(self.parameter_mapping)
#                 out_data = [{"is_set": False, "data": None} for _ in range(num_elems)]
#                 out_param_map = list(range(next_dat_idx, next_dat_idx + num_elems))
#                 self.parameter_data.extend(out_data)
#                 self.parameter_mapping.append(out_param_map)
#                 output_map_indices[output.typ] = next_map_idx

#         for i_idx, i in enumerate(init_multi):
#             element_indices.append(len(self.elements))
#             self.elements.append(
#                 {
#                     "inputs": [
#                         {
#                             "path": k,
#                             "parameter_mapping_index": input_map_indices[tuple(k)],
#                             "data_index": v,
#                         }
#                         for k, v in i["value_index"].items()
#                     ],
#                     "outputs": [
#                         {
#                             "path": ("outputs", k),
#                             "parameter_mapping_index": v,
#                             "data_index": i_idx,
#                         }
#                         for k, v in output_map_indices.items()
#                     ],
#                 }
#             )

#         self.element_indices.append(element_indices)
#         self.name_repeat_indices.append(
#             sum(i.template.name == task_template.name for i in self.tasks) + 1
#         )
#         task = Task(task_template, self, len(self.tasks))
#         self.tasks.add_object(task)

#     @staticmethod
#     def resolve_initial_elements(multi):
#         """
#         Parameters
#         ----------
#         multi : list of dict
#             Each list item represents a sequence of values with keys:
#                 multiplicity: int
#                 nesting_order: int
#                 address : str
#         """

#         # order by nesting order (so lower nesting orders will be fastest-varying):
#         multi_srt = sorted(multi, key=lambda x: x["nesting_order"])
#         multi_srt_grp = group_by_dict_key_values(multi_srt, "nesting_order")

#         elements = [{"value_index": {}}]
#         for para_sequences in multi_srt_grp:

#             # check all equivalent nesting_orders have equivalent multiplicities
#             all_multis = {i["multiplicity"] for i in para_sequences}
#             if len(all_multis) > 1:
#                 raise ValueError(
#                     f"All sequences with the same `nesting_order` must have the same "
#                     f"multiplicity, but found multiplicities {list(all_multis)!r} for "
#                     f"`nesting_order` of {para_sequences[0]['nesting_order']}."
#                 )

#             new_elements = []
#             for val_idx in range(para_sequences[0]["multiplicity"]):
#                 for element in elements:
#                     new_elements.append(
#                         {
#                             "value_index": {
#                                 **element["value_index"],
#                                 **{i["address"]: val_idx for i in para_sequences},
#                             }
#                         }
#                     )
#             elements = new_elements

#         return elements

#     def add_task_after(self, task):
#         pass

#     def add_task_before(self, task):
#         pass

#     def remove_task(self, task):
#         pass

#     def get_input_values(self, task_index, parameter_path):
#         """Get the value of an input for each element in a task."""
#         return [
#             self.get_input_value(
#                 task_index=task_index, element_index=i, parameter_path=parameter_path
#             )
#             for i in range(self.tasks[task_index].num_elements)
#         ]

#     def get_input_value(self, task_index, element_index, parameter_path):

#         element = self.elements[self.tasks[task_index].element_indices[element_index]]
#         current_value = None
#         for input_i in element["inputs"]:

#             param_data_idx = self.parameter_mapping[input_i["parameter_mapping_index"]][
#                 input_i["data_index"]
#             ]

#             is_parent = False
#             is_update = False
#             try:
#                 rel_path_parts = get_relative_path(parameter_path, input_i["path"])
#                 is_parent = True
#             except ValueError:
#                 try:
#                     update_path = get_relative_path(input_i["path"], parameter_path)
#                     is_update = True
#                 except ValueError:
#                     pass

#             if is_parent:
#                 # replace current value:
#                 final_data_path = (param_data_idx, "data", *rel_path_parts)
#                 try:
#                     current_value = get_in_container(
#                         self.parameter_data, final_data_path
#                     )  # or use Zarr to get from persistent
#                 except TypeError:
#                     # import traceback

#                     # traceback.print_exc()
#                     pass

#             elif is_update:
#                 # update sub-part of current value
#                 update_data = self.parameter_data[param_data_idx][
#                     "data"
#                 ]  # or use Zarr to get from persistent
#                 set_in_container(current_value, update_path, update_data)

#         return current_value

#     def make_workflow(self, path):
#         # TODO: make the workflow and save to path
#         wk = Workflow(path)
#         return wk

#     @classmethod
#     def from_spec(cls, spec, all_schemas, all_parameters):

#         # initialise task templates:
#         tasks = []
#         for i in spec.pop("tasks"):
#             tasks.append(TaskTemplate.from_spec(i, all_schemas, all_parameters))
#         spec["task_templates"] = tasks

#         return cls(**spec)


@dataclass
class WorkflowTemplate(JSONLike):
    """Class to represent initial parametrisation of a workflow, with limited validation
    logic."""

    _child_objects = (
        ChildObjectSpec(
            name="tasks",
            class_name="Task",
            is_multiple=True,
            parent_ref="workflow_template",
        ),
    )

    name: str
    tasks: Optional[List[Task]] = field(default_factory=lambda: [])
    workflow: Optional[Workflow] = None

    def __post_init__(self):
        self._set_parent_refs()

    @classmethod
    def from_YAML_file(cls, path):
        dat = read_YAML_file(path)
        cls.app._ensure_data_files()  # TODO: fix this at App
        return cls.from_json_like(dat, shared_data=cls.app.app_data)


class Workflow:
    """Class to represent a persistent workflow."""

    _app_attr = "app"

    def __init__(self, path):
        """Load a persistent workflow from a path."""

        self.path = path

        root = self._get_workflow_root_group(mode="r")

        self._persistent_metadata = root.attrs.asdict()

        self._shared_data = None
        self._tasks = None
        self._elements = None
        self._template = None

        self.history = root.attrs["history"]

    def _get_workflow_root_group(self, mode):
        try:
            return zarr.open(self.path, mode=mode)
        except zarr.errors.PathNotFoundError:
            raise WorkflowNotFoundError(
                f"No workflow found at path: {self.path}"
            ) from None

    @property
    def shared_data(self):
        if not self._shared_data:
            self._shared_data = self.app.shared_data_from_json_like(
                self._persistent_metadata["shared_data"]
            )
        return self._shared_data

    @property
    def template(self):
        if not self._template:
            self._template = self.app.WorkflowTemplate.from_json_like(
                self._persistent_metadata["template"],
                self.shared_data,
            )
            self._template.workflow = self

        return self._template

    @property
    def tasks(self):
        if self._tasks is None:
            self._tasks = WorkflowTaskList(
                [
                    WorkflowTask(
                        workflow=self, template=self.template.tasks[idx], index=idx, **i
                    )
                    for idx, i in enumerate(self._persistent_metadata["tasks"])
                ]
            )
        return self._tasks

    @property
    def num_tasks(self):
        return len(self._persistent_metadata["tasks"])

    @property
    def num_elements(self):
        return len(self._persistent_metadata["elements"])

    @property
    def elements(self):
        if not self._elements:
            self._elements = [
                Element(task=task, data_index=self._persistent_metadata["elements"][i])
                for task in self.tasks
                for i in task.element_indices
            ]
        return self._elements

    @property
    def task_name_repeat_idx(self):
        return self._persistent_metadata["task_name_repeat_idx"]

    @classmethod
    def _make_empty_workflow(
        cls,
        template: WorkflowTemplate,
        path=None,
        name=None,
        overwrite=False,
    ):
        """Generate a task-less workflow from a WorkflowTemplate, in preparation for
        adding valid tasks."""

        # Write initial Zarr root group and attributes, then add tasks/elements
        # incrementally:

        cls.app._ensure_data_files()  # TODO: fix this at App

        timestamp = datetime.now(timezone.utc).astimezone()
        history = {
            "timestamp_format": TS_FMT,
            "events": [
                {
                    "type": "create",
                    "at": timestamp.strftime(TS_FMT),
                    "machine": cls.app.config.get("machine"),
                }
            ],
        }

        path = Path(path or "").resolve()
        name = name or f"{template.name}_{timestamp.strftime(TS_NAME_FMT)}"
        path = path.joinpath(name)

        # TODO: extra out input values from template and save in zarr parameter data

        template_js, template_sh = template.to_json_like()
        # print(f"template_js: {template_js}")
        # print(f"template_sh: {template_sh}")

        root_attrs = {
            "history": history,
            "shared_data": template_sh,
            "template": template_js,
            "parameter_mapping": [],
            "elements": [],
            "tasks": [],
            "task_name_repeat_idx": [],
        }

        # TODO: intermittent Dropbox permission error; wrap in exception are retry?
        store = zarr.DirectoryStore(path)
        root = zarr.group(store=store, overwrite=overwrite)
        root.attrs.update(root_attrs)

        root.create_group("parameter_data")

        return cls.load(path)

    @classmethod
    def from_template(cls, template, path=None, name=None, overwrite=False):
        tasks = template.__dict__.pop("tasks") or []
        template.tasks = []
        obj = cls._make_empty_workflow(template, path, name, overwrite)
        for task in tasks:
            obj.add_task(task)
        return obj

    @classmethod
    def from_YAML_file(cls, YAML_path, path=None, name=None, overwrite=False):
        template = cls.app.WorkflowTemplate.from_YAML_file(YAML_path)
        return cls.from_template(template, path, name, overwrite)

    @classmethod
    def load(cls, path):
        """Alias for object initialisation."""
        return cls(path)

    def _resolve_input_source_task_reference(
        self, input_source: InputSource, new_task_name: str
    ):
        """Normalise the input source task reference and convert a source to a local type
        if required."""

        # TODO: test thoroughly!

        if isinstance(input_source.task_ref, str):
            if input_source.task_ref == new_task_name:
                if input_source.task_source_type is self.app.TaskSourceType.OUTPUT:
                    raise InvalidInputSourceTaskReference(
                        f"Input source {input_source.to_string()!r} cannot refer to the "
                        f"outputs of its own task!"
                    )
                else:
                    warn(
                        f"Changing input source {input_source.to_string()!r} to a local "
                        f"type, since the input source task reference refers to its own "
                        f"task."
                    )
                    # TODO: add an InputSource source_type setter to reset task_ref/source_type
                    input_source.source_type = self.app.InputSourceType.LOCAL
                    input_source.task_ref = None
                    input_source.task_source_type = None
            else:
                try:
                    uniq_names_cur = self.get_task_unique_names(map_to_insert_ID=True)
                    input_source.task_ref = uniq_names_cur[input_source.task_ref]
                except KeyError:
                    raise InvalidInputSourceTaskReference(
                        f"Input source {input_source.to_string()!r} refers to a missing "
                        f"or inaccessible task: {input_source.task_ref!r}."
                    )

    def ensure_input_sources(self, new_task: Task, new_index: int, new_name: str):
        """Check valid input sources are specified for a new task to be added to the
        workflow in a given position. If none are specified, set them according to the
        default behaviour."""

        # TODO: order sources by preference so can just take first in the case of input
        # source mode AUTO?
        available_sources = new_task.get_available_task_input_sources(
            self.template.tasks[:new_index]
        )

        # TODO: get available input sources from workflow imports

        print("available_sources")
        pprint(available_sources)

        # check any specified sources are valid:
        for schema_input in new_task.all_schema_inputs:
            for specified_source in new_task.input_sources.get(schema_input.typ, []):
                self._resolve_input_source_task_reference(specified_source, new_name)
                if specified_source not in available_sources[schema_input.typ]:
                    raise ValueError(
                        f"The following input source is not available for schema input "
                        f"{schema_input!r}: {specified_source.to_string()!r}. Available "
                        f"input sources are: "
                        f"{[i.to_string() for i in available_sources[schema_input.typ]]}"
                    )

        # print(f"\nall_sources:")
        # pprint(all_sources)

        print(f"new_task.unsourced_inputs")
        pprint(new_task.unsourced_inputs)

        # if an input is not specified at all in the `inputs` dict (what about when list?),
        # then check if there is an input files entry for associated inputs,
        # if there is

        # set source for any unsourced inputs:
        for input_type in new_task.unsourced_inputs:

            inp_i_sources = available_sources[input_type]
            print(f"inp_i_sources: {inp_i_sources}")
            # TODO: pre-sort available sources so that we can just pop off the first one?

            # input may not be required

            # # set source for this input:
            # if inp_i_sources["has_local"]:
            #     new_sources = [InputSource("local")]

            # elif inp_i_sources["tasks"]:
            #     # we can only take parameters with implicit propagation mode:
            #     params_info = []
            #     for (task_idx, task_name), params in inp_i_sources["tasks"].items():
            #         for i in params:
            #             if i.propagation_mode == ParameterPropagationMode.IMPLICIT:
            #                 params_info.append((i.input_or_output, task_idx, task_name))

            #     # sort by output/input (output first), then by task index (highest first):
            #     params_info = sorted(
            #         params_info, key=itemgetter(0, 1), reverse=True
            #     )  # TODO: TEST THIS!!!
            #     print(f"params_info: {params_info}")
            #     new_sources = [
            #         InputSource(f"tasks.{params_info[0][2]}.{params_info[0][0]}s")
            #     ]

            # else:
            #     # input may not need defining (if all associated input files are passed
            #     # and the input does not appear in any action commands)
            #     new_sources = None

            # new_task.input_sources.update({input_type: new_sources})

    def _dump_persistent_metadata(self):

        # invalidate workflow attributes to force re-init on access:
        self._tasks = None
        self._elements = None
        self._template = None
        self._shared_data = None
        self._template = None

        root = self._get_workflow_root_group(mode="r+")
        root.attrs.put(self._persistent_metadata)

    def get_zarr_parameter_group(self, group_idx):
        root = self._get_workflow_root_group(mode="r")
        return root.get(f"parameter_data/{group_idx}")

    @staticmethod
    def resolve_element_data_indices(multiplicities):
        """Find the index of the Zarr parameter group index list corresponding to each
        input data for all elements.

        Parameters
        ----------
        multiplicities : list of dict
            Each list item represents a sequence of values with keys:
                multiplicity: int
                nesting_order: int
                address : str

        Returns
        -------
        element_dat_idx : list of dict
            Each list item is a dict representing a single task element and whose keys are
            input data paths and whose values are indices that index the values of the
            dict returned by the `task.make_persistent` method.

        """

        # order by nesting order (so lower nesting orders will be fastest-varying):
        multi_srt = sorted(multiplicities, key=lambda x: x["nesting_order"])
        multi_srt_grp = group_by_dict_key_values(
            multi_srt, "nesting_order"
        )  # TODO: is tested?

        element_dat_idx = [{}]
        for para_sequences in multi_srt_grp:

            # check all equivalent nesting_orders have equivalent multiplicities
            all_multis = {i["multiplicity"] for i in para_sequences}
            if len(all_multis) > 1:
                raise ValueError(
                    f"All sequences with the same `nesting_order` must have the same "
                    f"multiplicity, but found multiplicities {list(all_multis)!r} for "
                    f"`nesting_order` of {para_sequences[0]['nesting_order']}."
                )

            new_elements = []
            for val_idx in range(para_sequences[0]["multiplicity"]):
                for element in element_dat_idx:
                    new_elements.append(
                        {
                            **element,
                            **{i["address"]: val_idx for i in para_sequences},
                        }
                    )
            element_dat_idx = new_elements

        return element_dat_idx

    def _add_parameter_group(self, data, is_set):

        root = self._get_workflow_root_group(mode="r+")
        param_dat_group = root.get("parameter_data")

        names = [int(i) for i in param_dat_group.keys()]
        new_idx = max(names) + 1 if names else 0
        new_name = str(new_idx)

        new_param_group = param_dat_group.create_group(name=new_name)
        zarr_encode(data, new_param_group)

        return new_idx

    def generate_new_elements(
        self, input_data_indices, output_data_indices, element_data_indices
    ):

        # print(f"input_data_indices: {input_data_indices}")
        # print(f"output_data_indices: {output_data_indices}")
        # print(f"element_data_indices: {element_data_indices}")

        new_elements = []
        for i_idx, i in enumerate(element_data_indices):
            elem_i = {k: input_data_indices[k][v] for k, v in i.items()}
            elem_i.update(
                {f"outputs.{k}": v[i_idx] for k, v in output_data_indices.items()}
            )
            # print(f"elem_i: {elem_i}")
            new_elements.append(elem_i)

        return new_elements

    def get_task_unique_names(self, map_to_insert_ID=False):
        """Return the unique names of all workflow tasks.

        Parameters
        ----------
        map_to_insert_ID : bool, optional
            If True, return a dict whose values are task insert IDs, otherwise return a
            list.

        """

        names = Task.get_task_unique_names(self.template.tasks)
        if map_to_insert_ID:
            insert_IDs = (i.insert_ID for i in self.template.tasks)
            return dict(zip(names, insert_IDs))
        else:
            return names

    def _get_new_task_unique_name(self, new_task, new_index):

        task_templates = list(self.template.tasks)
        task_templates.insert(new_index, new_task)
        uniq_names = Task.get_task_unique_names(task_templates)

        return uniq_names[new_index]

    def add_task(self, task: Task, new_index=None):

        print(f"task.resources: {task.resources}")

        if new_index is None:
            new_index = self.num_tasks

        new_task_name = self._get_new_task_unique_name(task, new_index)
        input_sources = self.ensure_input_sources(task, new_index, new_task_name)

        print("input_sources")
        pprint(input_sources)

        # TODO: use input sources!

        # TODO: also save the original Task object, since it may be modified before
        # adding to the workflow

        # TODO: think about ability to change input sources? (in the context of adding a new task)
        # what happens when new insert a new task and it outputs a parameter that is used by a downstream task?
        # does it depend on what we originally specify as the downstream tasks's input sources?

        # add input source type "auto" (which is set by default)
        # then a "resolved source" property which is resolved to a specific source depending on the available sources?

        # TODO: perhaps refactor into an `add_elements` that can also be used to add
        # elements to a pre-existing task?

        input_data_idx = task.make_persistent(self)
        multiplicities = task.prepare_element_resolution()
        element_data_idx = self.resolve_element_data_indices(multiplicities)

        output_data_idx = task._prepare_persistent_outputs(self, len(element_data_idx))
        elements = self.generate_new_elements(
            input_data_idx, output_data_idx, element_data_idx
        )

        element_indices = list(
            range(len(self.elements), len(self.elements) + len(elements))
        )

        # print("input_data_indices")
        # pprint(input_data_idx)

        # print("multiplicities")
        # pprint(multiplicities)

        # print("element_data_indices")
        # pprint(element_data_idx)

        # print("elements")
        # pprint(elements)

        task._insert_ID = self.num_tasks
        task._dir_name = f"task_{task.insert_ID}_{new_task_name}"

        task_js, task_shared_data = task.to_json_like()

        # add any missing shared data for this task template:
        for shared_name, shared_data in task_shared_data.items():
            if shared_name not in self._persistent_metadata["shared_data"]:
                self._persistent_metadata["shared_data"][shared_name] = {}
            for k, v in shared_data.items():
                if k not in self._persistent_metadata["shared_data"][shared_name]:
                    self._persistent_metadata["shared_data"][shared_name][k] = v

        self._persistent_metadata["template"]["tasks"].append(task_js)

        wk_task = {"element_indices": element_indices}

        self._persistent_metadata["tasks"].insert(new_index, wk_task)
        self._persistent_metadata["elements"].extend(elements)

        self._dump_persistent_metadata()

    def add_task_after(self, task_ref):
        # TODO: find position of new task, then call add_task
        # TODO: add new downstream elements?
        pass

    def add_task_before(self, task_ref):
        # TODO: find position of new task, then call add_task
        # TODO: add new downstream elements?
        pass

    def rename(self, new_name):
        pass

    def add_submission(self, filter):
        pass


@dataclass
class WorkflowBlueprint:
    """Pre-built workflow templates that are simpler to parametrise (e.g. fitting workflows)."""

    workflow_template: WorkflowTemplate



from __future__ import annotations
from dataclasses import dataclass, field
import enum
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np

from hpcflow.sdk.core.errors import ValuesAlreadyPersistentError
from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike
from hpcflow.sdk.core.utils import check_valid_py_identifier
from hpcflow.sdk.core.zarr_io import ZarrEncodable, zarr_decode


Address = List[Union[int, float, str]]
Numeric = Union[int, float, np.number]


class ParameterPropagationMode(enum.Enum):

    IMPLICIT = 0
    EXPLICIT = 1
    NEVER = 2


@dataclass
class ParameterPath(JSONLike):

    path: Sequence[Union[str, int, float]]
    task: Optional[Union[TaskTemplate, TaskSchema]] = None  # default is "current" task


@dataclass
class Parameter(JSONLike):

    _validation_schema = "parameters_spec_schema.yaml"
    _child_objects = (
        ChildObjectSpec(
            name="typ",
            json_like_name="type",
        ),
    )

    typ: str
    is_file: bool = False
    sub_parameters: List[SubParameter] = field(default_factory=lambda: [])
    _value_class: Any = None
    _hash_value: Optional[str] = field(default=None, repr=False)

    def __post_init__(self):
        self.typ = check_valid_py_identifier(self.typ)
        for i in ZarrEncodable.__subclasses__():
            if i._typ == self.typ:
                self._value_class = i

    def to_dict(self):
        dct = super().to_dict()
        del dct["_value_class"]
        return dct


@dataclass
class SubParameter:
    address: Address
    parameter: Parameter


@dataclass
class SchemaParameter(JSONLike):

    _child_objects = (
        ChildObjectSpec(
            name="parameter",
            class_name="SchemaInput",
            shared_data_name="parameters",
            shared_data_primary_key="typ",
        ),
    )

    def __post_init__(self):
        self._validate()

    def _validate(self):
        if isinstance(self.parameter, str):
            self.parameter = Parameter(self.parameter)

    @property
    def name(self):
        return self.parameter.name

    @property
    def typ(self):
        return self.parameter.typ


@dataclass
class SchemaInput(SchemaParameter):
    """A Parameter as used within a particular schema, for which a default value may be
    applied."""

    _child_objects = (
        ChildObjectSpec(
            name="parameter",
            class_name="SchemaInput",
            shared_data_name="parameters",
            shared_data_primary_key="typ",
        ),
        ChildObjectSpec(
            name="default_value",
            class_name="InputValue",
        ),
        ChildObjectSpec(
            name="propagation_mode",
            class_name="ParameterPropagationMode",
            is_enum=True,
        ),
    )

    parameter: Parameter
    default_value: Optional[InputValue] = None
    propagation_mode: ParameterPropagationMode = ParameterPropagationMode.IMPLICIT

    # can we define elements groups on local inputs as well, or should these be just for
    # elements from other tasks?
    group: Optional[str] = None
    where: Optional[ElementFilter] = None

    def _validate(self):
        super()._validate()
        if self.default_value is not None:
            if not isinstance(self.default_value, self.app.InputValue):
                self.default_value = self.app.InputValue(
                    parameter=self.parameter,
                    value=self.default_value,
                )
            if self.default_value.parameter != self.parameter:
                raise ValueError(
                    f"{self.__class__.__name__} `default_value` must be an `InputValue` for "
                    f"parameter: {self.parameter!r}, but specified `InputValue` parameter "
                    f"is: {self.default_value.parameter!r}."
                )

    @property
    def input_or_output(self):
        return "input"

    # @classmethod
    # def from_json_like(cls, json_like):
    #     kwargs = cls.prepare_from_json_like(json_like)
    #     cls.app.logger.warn(f"SchemaInput.from_json_like: kwargs: {kwargs}")
    #     return super().from_json_like(kwargs)

    # @classmethod
    # @required_keys(1, "parameter")
    # @allowed_keys(1, "parameter", "default_value", "propagation_mode")
    # @check_in_object_list(spec_name="parameter")
    # def from_spec(cls, spec, parameters):
    #     cls.app.logger.debug("SchemaInput.from_spec")
    #     if "default_value" in spec:
    #         spec["default_value"] = InputValue(
    #             parameter=parameters[spec["parameter"]], value=spec["default_value"]
    #         )
    #     return super().from_spec(spec, parameters)


@dataclass
class SchemaOutput(SchemaParameter):
    """A Parameter as outputted from particular task."""

    parameter: Parameter
    propagation_mode: ParameterPropagationMode = ParameterPropagationMode.IMPLICIT

    @property
    def input_or_output(self):
        return "output"


@dataclass
class BuiltinSchemaParameter:
    # builtin inputs (resources,parameter_perturbations,method,implementation
    # builtin outputs (time, memory use, node/hostname etc)
    # - builtin parameters do not propagate to other tasks (since all tasks define the same
    #   builtin parameters).
    # - however, builtin parameters can be accessed if a downstream task schema specifically
    #   asks for them (e.g. for calculating/plotting a convergence test)
    pass


class ValueSequence(JSONLike):
    def __init__(
        self,
        path: str,
        nesting_order: int,
        values: List[Any],
    ):
        self.path = path
        self.nesting_order = nesting_order
        self._values = values

        self._values_group_idx = None
        self._workflow = None

    def __repr__(self):
        vals_grp_idx = (
            f"values_group_idx={self._values_group_idx}, "
            if self._values_group_idx
            else ""
        )
        return (
            f"{self.__class__.__name__}("
            f"path={self.path!r}, "
            f"nesting_order={self.nesting_order}, "
            f"{vals_grp_idx}"
            f"values={self.values}"
            f")"
        )

    @classmethod
    def _json_like_constructor(cls, json_like):
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""

        _values_group_idx = json_like.pop("_values_group_idx", None)
        if "_values" in json_like:
            json_like["values"] = json_like.pop("_values")

        obj = cls(**json_like)
        obj._values_group_idx = _values_group_idx
        return obj

    def to_dict(self):
        out = super().to_dict()
        if "_workflow" in out:
            del out["_workflow"]
        return out

    def check_address_exists(self, value):
        """Check a given nested dict/list "address" resolves to somewhere within
        `value`."""
        if self.address:
            sub_val = value
            for i in self.address:
                try:
                    sub_val = sub_val[i]
                except (IndexError, KeyError, TypeError):
                    msg = (
                        f"Address {self.address} does not exist in the base "
                        f"value: {value}"
                    )
                    raise ValueError(msg)

    def _get_param_path(self):
        return self.path  # TODO: maybe not needed?

    def make_persistent(self, workflow):
        """Save value to a persistent workflow."""

        # TODO: test raise
        if self._values_group_idx is not None:
            raise ValuesAlreadyPersistentError(
                f"{self.__class__.__name__} is already persistent."
            )

        param_group_idx = []
        for i in self._values:
            pg_idx_i = workflow._add_parameter_group(i, is_set=True)
            param_group_idx.append(pg_idx_i)

        self._values_group_idx = param_group_idx
        self._workflow = workflow
        self._values = None
        return {self._get_param_path(): param_group_idx}

    @property
    def workflow(self):
        return self._workflow

    @property
    def values(self):
        if self._values_group_idx is not None:
            vals = []
            for pg_idx_i in self._values_group_idx:
                grp = self._workflow.get_zarr_parameter_group(pg_idx_i)
                vals.append(zarr_decode(grp))
            return vals
        else:
            return self._values

    @classmethod
    def from_linear_space(cls, start, stop, num=50, address=None, **kwargs):
        values = list(np.linspace(start, stop, num=num, **kwargs))
        return cls(values, address=address)

    @classmethod
    def from_range(cls, start, stop, step=1, address=None):
        if isinstance(step, int):
            return cls(values=list(np.arange(start, stop, step)), address=address)
        else:
            # Use linspace for non-integer step, as recommended by Numpy:
            return cls.from_linear_space(
                start,
                stop,
                num=int((stop - start) / step),
                address=address,
                endpoint=False,
            )


@dataclass
class AbstractInputValue(JSONLike):
    """Class to represent all sequence-able inputs to a task."""

    _workflow = None

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"_value_group_idx={self._value_group_idx}, "
            f"value={self.value}"
            f")"
        )

    def to_dict(self):
        out = super().to_dict()
        if "_workflow" in out:
            del out["_workflow"]
        return out

    def make_persistent(self, workflow) -> Dict:
        """Save value to a persistent workflow.

        Parameters
        ----------
        workflow : Workflow

        Returns
        -------
        dict of tuple : int
            Single-item dict whose key is the data path for this task input and whose
            value is the integer index of the parameter data Zarr group where the data is
            stored.
        """

        param_group_idx = workflow._add_parameter_group(self._value, is_set=True)
        self._value_group_idx = param_group_idx
        self._workflow = workflow
        self._value = None
        return {self._get_param_path(): [param_group_idx]}

    @property
    def workflow(self):
        if self._workflow:
            return self._workflow
        elif self._task:
            return self._task.workflow_template.workflow

    @property
    def value(self):
        if self._value_group_idx is not None:
            grp = self.workflow.get_zarr_parameter_group(self._value_group_idx)
            val = zarr_decode(grp)
            if self.parameter._value_class:
                val = self.parameter._value_class(**val)
        else:
            val = self._value

        return val


@dataclass
class ValuePerturbation(AbstractInputValue):
    name: str
    path: Optional[Sequence[Union[str, int, float]]] = None
    multiplicative_factor: Optional[Numeric] = 1
    additive_factor: Optional[Numeric] = 0

    @classmethod
    def from_spec(cls, spec):
        return cls(**spec)


class InputValue(AbstractInputValue):

    _child_objects = (
        ChildObjectSpec(
            name="parameter",
            class_name="SchemaInput",
            shared_data_primary_key="typ",
            shared_data_name="parameters",
        ),
    )

    def __init__(
        self,
        parameter: Union[Parameter, SchemaInput, str],
        value: Optional[Any] = None,
        path: Optional[str] = None,
    ):
        if isinstance(parameter, str):
            parameter = self.app.parameters.get(parameter)

        self.parameter = parameter
        self.path = (path.strip(".") if path else None) or None
        self._value = value

        self._value_group_idx = None  # assigned by method make_persistent
        self._task = None  # assigned by parent Task

    def __repr__(self):

        val_grp_idx = ""
        if self._value_group_idx is not None:
            val_grp_idx = f", value_group_idx={self._value_group_idx}"

        path_str = ""
        if self.path is not None:
            path_str = f", path={self.path!r}"

        return (
            f"{self.__class__.__name__}("
            f"parameter={self.parameter.typ!r}, "
            f"value={self.value}"
            f"{path_str}"
            f"{val_grp_idx}"
            f")"
        )

    @classmethod
    def _json_like_constructor(cls, json_like):
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""

        _value_group_idx = json_like.pop("_value_group_idx", None)
        if "_value" in json_like:
            json_like["value"] = json_like.pop("_value")

        obj = cls(**json_like)
        obj._value_group_idx = _value_group_idx

        return obj

    def _get_param_path(self):
        return f"inputs.{self.parameter.typ}" f"{f'.{self.path}' if self.path else ''}"

    @classmethod
    def from_json_like(cls, json_like, shared_data=None):

        if "path" not in json_like:
            param_spec = json_like["parameter"].split(".")
            json_like["parameter"] = param_spec[0]
            json_like["path"] = ".".join(param_spec[1:])

        obj = super().from_json_like(json_like, shared_data)

        return obj

    @property
    def is_sub_value(self):
        """True if the value is for a sub part of the parameter (i.e. if `path` is set).
        Sub-values are not added to the base parameter data, but are interpreted as
        single-value sequences."""
        return True if self.path else False


class ResourceSpec(JSONLike):

    _resource_list = None

    _child_objects = (
        ChildObjectSpec(
            name="scope",
            class_name="ActionScope",
        ),
    )

    def __init__(self, scope=None, scratch=None, num_cores=None):
        self.scope = scope or self.app.ActionScope.any()
        self.scratch = scratch
        self.num_cores = num_cores

    def __repr__(self):
        scratch_str = ""
        if self.scratch is not None:
            scratch_str = f", scratch={self.scratch!r}"
        num_cores_str = ""
        if self.num_cores is not None:
            num_cores_str = f", num_cores={self.num_cores!r}"

        return (
            f"{self.__class__.__name__}("
            f"scope={self.scope}"
            f"{scratch_str}"
            f"{num_cores_str}"
            f")"
        )

    def _get_param_path(self):
        scope_str = ""
        if self.scope.typ.name != self.app.ActionScopeType.ANY.name:
            scope_str = f".{self.scope.to_string()}"
        return f"resources{scope_str}"

    @property
    def task(self):
        return self._resource_list.task


class InputSourceType(enum.Enum):

    IMPORT = 0
    LOCAL = 1
    DEFAULT = 2
    TASK = 3


class TaskSourceType(enum.Enum):
    INPUT = 0
    OUTPUT = 1
    ANY = 2


class InputSourceMode(enum.Enum):
    """Set to MANUAL if a task has input source(s) specified on creation (or modification)
    otherwise set to AUTO, in which case input sources will be set by hpcflow, and input
    sources may be appended to if new tasks/imports are added to the workflow."""

    AUTO = 0
    MANUAL = 1


class InputSource(JSONLike):

    _child_objects = (
        ChildObjectSpec(
            name="source_type",
            json_like_name="type",
            class_name="InputSourceType",
            is_enum=True,
        ),
    )

    def __init__(
        self,
        source_type,
        import_ref=None,
        task_ref=None,
        task_source_type=None,
        where=None,
    ):

        self.source_type = self._validate_source_type(source_type)
        self.import_ref = import_ref
        self.task_ref = task_ref
        self.task_source_type = self._validate_task_source_type(task_source_type)
        self.where = where

        if self.source_type is InputSourceType.TASK:
            if self.task_ref is None:
                raise ValueError(f"Must specify `task_ref` if `source_type` is TASK.")
            if self.task_source_type is None:
                self.task_source_type = TaskSourceType.OUTPUT

        if self.source_type is InputSourceType.IMPORT and self.import_ref is None:
            raise ValueError(f"Must specify `import_ref` if `source_type` is IMPORT.")

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        elif (
            self.source_type == other.source_type
            and self.import_ref == other.import_ref
            and self.task_ref == other.task_ref
            and self.task_source_type == other.task_source_type
            and self.where == other.where
        ):
            return True
        else:
            return False

    def __repr__(self) -> str:
        members = {k: v for k, v in self.__dict__.items() if v is not None}
        return (
            f"{self.__class__.__name__}("
            f'{", ".join([f"{k}={repr(v)}" for k, v in members.items()])}'
            f")"
        )

    def to_string(self):
        out = [self.source_type.name.lower()]
        if self.source_type is InputSourceType.TASK:
            out += [str(self.task_ref), self.task_source_type.name.lower()]
        elif self.source_type is InputSourceType.IMPORT:
            out += [str(self.import_ref)]
        return ".".join(out)

    @staticmethod
    def _validate_source_type(src_type):
        if src_type is None:
            return None
        if isinstance(src_type, InputSourceType):
            return src_type
        try:
            src_type = getattr(InputSourceType, src_type.upper())
        except AttributeError:
            raise ValueError(
                f"InputSource `source_type` specified as {src_type!r}, but "
                f"must be one of: {[i.name for i in InputSourceType]!r}."
            )
        return src_type

    @staticmethod
    def _validate_task_source_type(task_src_type):
        if task_src_type is None:
            return None
        if isinstance(task_src_type, TaskSourceType):
            return task_src_type
        try:
            task_source_type = getattr(TaskSourceType, task_src_type.upper())
        except AttributeError:
            raise ValueError(
                f"InputSource `task_source_type` specified as {task_src_type!r}, but "
                f"must be one of: {[i.name for i in TaskSourceType]!r}."
            )
        return task_source_type

    @classmethod
    def from_string(cls, str_defn):
        return cls(**cls._parse_from_string(str_defn))

    @classmethod
    def _parse_from_string(cls, str_defn):
        """Parse a dot-delimited string definition of an InputSource.

        Examples:
            - task.[task_ref].input
            - task.[task_ref].output
            - local
            - default
            - import.[import_ref]

        """
        parts = str_defn.lower().split(".")
        source_type = cls._validate_source_type(parts[0])
        task_ref = None
        task_source_type = None
        import_ref = None
        if (
            (
                source_type in (InputSourceType.LOCAL, InputSourceType.DEFAULT)
                and len(parts) > 1
            )
            or (source_type is InputSourceType.TASK and len(parts) > 3)
            or (source_type is InputSourceType.IMPORT and len(parts) > 2)
        ):
            raise ValueError(f"InputSource string not understood: {str_defn!r}.")

        if source_type is InputSourceType.TASK:
            task_ref = parts[1]
            try:
                task_ref = int(task_ref)
            except ValueError:
                pass
            try:
                task_source_type_str = parts[2]
            except IndexError:
                task_source_type_str = TaskSourceType.OUTPUT
            task_source_type = cls._validate_task_source_type(task_source_type_str)
        elif source_type is InputSourceType.IMPORT:
            import_ref = parts[1]
            try:
                import_ref = int(import_ref)
            except ValueError:
                pass

        return {
            "source_type": source_type,
            "import_ref": import_ref,
            "task_ref": task_ref,
            "task_source_type": task_source_type,
        }

    @classmethod
    def from_json_like(cls, json_like, shared_data=None):
        if isinstance(json_like, str):
            json_like = cls._parse_from_string(json_like)
        return super().from_json_like(json_like, shared_data)

    @classmethod
    def import_(cls, import_ref):
        return cls(source_type=InputSourceType.IMPORT, import_ref=import_ref)

    @classmethod
    def local(cls):
        return cls(source_type=InputSourceType.LOCAL)

    @classmethod
    def default(cls):
        return cls(source_type=InputSourceType.DEFAULT)

    @classmethod
    def task(cls, task_ref, task_source_type=None):
        if not task_source_type:
            task_source_type = TaskSourceType.OUTPUT
        return cls(
            source_type=InputSourceType.TASK,
            task_ref=task_ref,
            task_source_type=cls._validate_task_source_type(task_source_type),
        )


@dataclass
class ParameterValue:
    @property
    def value(self):
        pass


@dataclass
class AvailableInputSources:
    """Container for listing the available sources for a given input within a task."""

    # TODO: should this be a list of `InputSource`?

    has_local: bool
    has_default: bool
    tasks: Dict
    imports: Dict



import pytest
from hpcflow.sdk.core.parameters import InputSource, InputSourceType, TaskSourceType


def test_input_source_class_method_local():
    assert InputSource.local() == InputSource(InputSourceType.LOCAL)


def test_input_source_class_method_default():
    assert InputSource.default() == InputSource(InputSourceType.DEFAULT)


def test_input_source_class_method_task():
    task_ref = 0
    assert InputSource.task(task_ref) == InputSource(
        source_type=InputSourceType.TASK, task_ref=task_ref
    )


def test_input_source_class_method_import():
    import_ref = (
        0  # TODO: interface to imports (and so how to reference) is not yet decided
    )
    assert InputSource.import_(import_ref) == InputSource(
        InputSourceType.IMPORT, import_ref=import_ref
    )


def test_input_source_class_method_task_same_default_task_source_type():
    task_ref = 0
    assert (
        InputSource(InputSourceType.TASK, task_ref=task_ref).task_source_type
        == InputSource.task(task_ref=task_ref).task_source_type
    )


def test_input_source_validate_source_type_string_local():
    assert InputSource("local") == InputSource(InputSourceType.LOCAL)


def test_input_source_validate_source_type_string_default():
    assert InputSource("default") == InputSource(InputSourceType.DEFAULT)


def test_input_source_validate_source_type_string_task():
    task_ref = 0
    assert InputSource("task", task_ref=task_ref) == InputSource(
        InputSourceType.TASK, task_ref=task_ref
    )


def test_input_source_validate_source_type_string_import():
    import_ref = (
        0  # TODO: interface to imports (and so how to reference) is not yet decided
    )
    assert InputSource("import", import_ref=import_ref) == InputSource(
        InputSourceType.IMPORT, import_ref=import_ref
    )


def test_input_source_validate_source_type_raise_on_unknown_string():
    with pytest.raises(ValueError):
        InputSource("bad_source_type")


def test_input_source_validate_task_source_type_string_any():
    task_ref = 0
    assert InputSource(
        InputSourceType.TASK, task_ref=task_ref, task_source_type="any"
    ) == InputSource(
        InputSourceType.TASK, task_ref=task_ref, task_source_type=TaskSourceType.ANY
    )


def test_input_source_validate_task_source_type_string_input():
    task_ref = 0
    assert InputSource(
        InputSourceType.TASK, task_ref=task_ref, task_source_type="input"
    ) == InputSource(
        InputSourceType.TASK, task_ref=task_ref, task_source_type=TaskSourceType.INPUT
    )


def test_input_source_validate_task_source_type_string_output():
    task_ref = 0
    assert InputSource(
        InputSourceType.TASK, task_ref=task_ref, task_source_type="output"
    ) == InputSource(
        InputSourceType.TASK, task_ref=task_ref, task_source_type=TaskSourceType.OUTPUT
    )


def test_input_source_validate_task_source_type_raise_on_unknown_string():
    task_ref = 0
    with pytest.raises(ValueError):
        InputSource(
            InputSourceType.TASK,
            task_ref=task_ref,
            task_source_type="bad_task_source_type",
        )


def test_input_source_to_string_local():
    assert InputSource.local().to_string() == "local"


def test_input_source_to_string_default():
    assert InputSource.default().to_string() == "default"


def test_input_source_to_string_task_output():
    task_ref = 0
    assert (
        InputSource.task(task_ref, task_source_type="output").to_string()
        == f"task.{task_ref}.output"
    )


def test_input_source_to_string_task_input():
    task_ref = 0
    assert (
        InputSource.task(task_ref, task_source_type="input").to_string()
        == f"task.{task_ref}.input"
    )


def test_input_source_to_string_task_any():
    task_ref = 0
    assert (
        InputSource.task(task_ref, task_source_type="any").to_string()
        == f"task.{task_ref}.any"
    )


def test_input_source_to_string_import():
    import_ref = 0
    assert InputSource.import_(import_ref).to_string() == f"import.{import_ref}"


def test_input_source_from_string_local():
    assert InputSource.from_string("local") == InputSource(InputSourceType.LOCAL)


def test_input_source_from_string_default():
    assert InputSource.from_string("default") == InputSource(InputSourceType.DEFAULT)


def test_input_source_from_string_task():
    assert InputSource.from_string("task.0.output") == InputSource(
        InputSourceType.TASK, task_ref=0, task_source_type=TaskSourceType.OUTPUT
    )


def test_input_source_from_string_task_same_default_task_source():
    task_ref = 0
    assert InputSource.from_string(f"task.{task_ref}") == InputSource(
        InputSourceType.TASK, task_ref=task_ref
    )


def test_input_source_from_string_import():
    import_ref = 0
    assert InputSource.from_string(f"import.{import_ref}") == InputSource(
        InputSourceType.IMPORT, import_ref=import_ref
    )



from PyInstaller.utils.hooks import collect_data_files

hiddenimports = ["hpcflow.sdk.data", "click.testing"]

datas = collect_data_files("hpcflow.sdk.data")
datas += collect_data_files(
    "hpcflow.tests",
    include_py_files=True,
    excludes=("**/__pycache__",),
)



from dataclasses import dataclass
from .parameters import Parameter

from valida.conditions import ConditionLike


@dataclass
class StoppingCriterion:
    parameter: Parameter
    condition: ConditionLike


@dataclass
class Loop:
    parameter: Parameter
    stopping_criteria: StoppingCriterion  # TODO: should be a logical combination of these (maybe provide a superclass in valida to re-use some logic there?)
    maximum_iterations: int



"""An hpcflow application."""

from functools import wraps
import warnings

import click
from colorama import init as colorama_init
from termcolor import colored

from hpcflow import __version__
from .core.json_like import JSONLike
from .core.utils import read_YAML_file
from . import api, SDK_logger
from .config import Config
from .config.cli import get_config_CLI
from .config.errors import ConfigError
from .core.actions import ActionScopeType
from .core.environment import Executable, NumCores
from .core.zarr_io import ZarrEncodable
from .core.parameters import (
    InputSourceMode,
    InputSourceType,
    ParameterPropagationMode,
    TaskSourceType,
    ValueSequence,
)
from .core.task import WorkflowTask
from .core.task_schema import TaskObjective
from .core.workflow import Workflow
from .log import AppLog
from .runtime import RunTimeInfo

SDK_logger = SDK_logger.getChild(__name__)


class BaseApp:
    """Class to generate the base hpcflow application."""

    def __init__(
        self,
        name,
        version,
        description,
        config_options,
        pytest_args=None,
    ):
        SDK_logger.info(f"Generating {self.__class__.__name__} {name!r}.")

        self.name = name
        self.version = version
        self.description = description
        self.config_options = config_options
        self.pytest_args = pytest_args

        self.CLI = self._make_CLI()
        self.log = AppLog(self)
        self.config = None
        self.run_time_info = RunTimeInfo(
            self.name, self.version, self.runtime_info_logger
        )

        # Set by `_load_data_files`:
        self._parameters = None
        self._command_files = None
        self._envs = None
        self._task_schemas = None
        self._app_data = {}

        self._core_classes = self._assign_core_classes()

        # Add API functions as methods:
        SDK_logger.debug(f"Assigning API functions to the {self.__class__.__name__}.")

        def get_api_method(func):
            # this function avoids scope issues
            return lambda *args, **kwargs: func(self, *args, **kwargs)

        all_funcs = [func_i for func_i in api.__dict__.values() if callable(func_i)]
        for func in all_funcs:

            if type(self) is BaseApp and func.__name__ == "run_hpcflow_tests":
                # this method provides the same functionality as the `run_tests` method
                continue

            SDK_logger.debug(f"Wrapping API callable: {func!r}")
            # allow sub-classes to override API functions:
            if not hasattr(self, func.__name__):
                api_method = get_api_method(func)
                api_method = wraps(func)(api_method)
                api_method.__doc__ = func.__doc__.format(name=name)
                setattr(self, func.__name__, api_method)

    def _get_core_JSONLike_classes(self):
        """Get all JSONLike subclasses (recursively)."""

        def all_subclasses(cls):
            return set(cls.__subclasses__()).union(
                [s for c in cls.__subclasses__() for s in all_subclasses(c)]
            )

        return all_subclasses(JSONLike)

    def inject_into(self, cls):
        SDK_logger.debug(f"Injecting app {self!r} into class {cls.__name__}")
        return type(cls.__name__, (cls,), {getattr(cls, "_app_attr"): self})

    def _assign_core_classes(self):

        core_classes = list(self._get_core_JSONLike_classes())

        # Non-`JSONLike` classes:
        core_classes += [
            ActionScopeType,
            Executable,
            InputSourceMode,
            InputSourceType,
            NumCores,
            ParameterPropagationMode,
            TaskObjective,
            TaskSourceType,
            ValueSequence,
            Workflow,
            WorkflowTask,
            ZarrEncodable,
        ]
        for cls in core_classes:
            if hasattr(cls, "_app_attr"):
                setattr(self, cls.__name__, self.inject_into(cls))
            else:
                setattr(self, cls.__name__, cls)

        return tuple(
            sorted(
                core_classes,
                key=lambda x: f"{x.__module__}.{x.__qualname__}",
            )
        )

    def _ensure_data_files(self):
        if not self.is_data_files_loaded:
            self._load_data_files()

    def _load_data_files(self):
        if not self.is_config_loaded:
            self.load_config()

        self._parameters = self._load_parameters()
        self._app_data["parameters"] = self._parameters

        self._command_files = self._load_command_files()
        self._app_data["command_files"] = self._command_files

        self._envs = self._load_environments()
        self._app_data["envs"] = self._envs

        self._task_schemas = self._load_task_schemas()
        self._app_data["task_schemas"] = self._task_schemas

        self.logger.info("Data files loaded.")

    def load_data_files(self):
        if self.is_data_files_loaded:
            warnings.warn("Data files already loaded; reloading.")
        self._load_data_files()

    def reload_data_files(self):
        if not self.is_data_files_loaded:
            warnings.warn("Data files not loaded; loading.")
        self._load_data_files()

    @property
    def app_data(self):
        return self._app_data

    @property
    def task_schemas(self):
        self._ensure_data_files()
        return self._task_schemas

    @property
    def parameters(self):
        self._ensure_data_files()
        return self._parameters

    @property
    def envs(self):
        self._ensure_data_files()
        return self._envs

    @property
    def command_files(self):
        self._ensure_data_files()
        return self._command_files

    @property
    def logger(self):
        return self.log.logger

    @property
    def API_logger(self):
        return self.logger.getChild("api")

    @property
    def CLI_logger(self):
        return self.logger.getChild("cli")

    @property
    def config_logger(self):
        return self.logger.getChild("config")

    @property
    def runtime_info_logger(self):
        return self.logger.getChild("runtime")

    @property
    def is_config_loaded(self):
        return bool(self.config)

    @property
    def is_data_files_loaded(self):
        return bool(self._parameters)

    def _load_config(self, config_dir, **overrides):
        self.logger.debug("Loading configuration.")
        self.config = Config(
            app=self,
            options=self.config_options,
            config_dir=config_dir,
            logger=self.config_logger,
            **overrides,
        )
        self.logger.info(f"Configuration loaded from: {self.config.config_file_path}")

    def load_config(self, config_dir=None, **overrides):
        if self.is_config_loaded:
            warnings.warn("Configuration is already loaded; reloading.")
        self._load_config(config_dir, **overrides)

    def reload_config(self, config_dir=None, **overrides):
        if not self.is_config_loaded:
            warnings.warn("Configuration is not loaded; loading.")
        self._load_config(config_dir, **overrides)

    def _make_API_CLI(self):
        """Generate the CLI for the main functionality."""

        @click.command(help=f"Generate a new {self.name} workflow")
        def make_workflow():
            self.make_workflow(dir=".")

        @click.command(help=f"Run {self.name} test suite.")
        def test():
            self.run_tests()

        @click.command(help=f"Run hpcflow test suite.")
        def test_hpcflow():
            self.run_hpcflow_tests()

        commands = [
            make_workflow,
            test,
        ]

        if type(self) is not BaseApp:
            # `test_hpcflow` is the same as `test` for the BaseApp
            commands.append(test_hpcflow)

        return commands

    def _make_CLI(self):
        """Generate the root CLI for the app."""

        colorama_init(autoreset=True)

        def run_time_info_callback(ctx, param, value):
            if not value or ctx.resilient_parsing:
                return
            click.echo(self.run_time_info)
            ctx.exit()

        @click.group(name=self.name)
        @click.version_option(
            version=self.version, package_name=self.name, prog_name=self.name
        )
        @click.version_option(
            __version__,
            "--hpcflow-version",
            help="Show the version of hpcflow and exit.",
            package_name="hpcflow",
            prog_name="hpcflow",
        )
        @click.help_option()
        @click.option(
            "--run-time-info",
            help="Print run-time information!",
            is_flag=True,
            is_eager=True,
            expose_value=False,
            callback=run_time_info_callback,
        )
        @click.option("--config-dir", help="Set the configuration directory.")
        @click.option(
            "--with-config",
            help="Override a config item in the config file",
            nargs=2,
            multiple=True,
        )
        @click.pass_context
        def new_CLI(ctx, config_dir, with_config):
            overrides = {kv[0]: kv[1] for kv in with_config}
            try:
                self.load_config(config_dir=config_dir, **overrides)
            except ConfigError as err:
                click.echo(f"{colored(err.__class__.__name__, 'red')}: {err}")
                ctx.exit(1)

        new_CLI.__doc__ = self.description
        new_CLI.add_command(get_config_CLI(self))
        for cli_cmd in self._make_API_CLI():
            new_CLI.add_command(cli_cmd)

        return new_CLI

    def _load_environments(self):

        all_envs = []
        for path in self.config.environment_files:
            all_envs.extend(read_YAML_file(path))

        return self.EnvironmentsList.from_json_like(all_envs)

    def _load_command_files(self):

        all_files = []
        for path in self.config.file_files:
            all_files.extend(read_YAML_file(path))

        return self.CommandFilesList.from_json_like(all_files)

    def _load_parameters(self):

        all_params = []
        for path in self.config.parameter_files:
            all_params.extend(read_YAML_file(path))

        return self.ParametersList.from_json_like(all_params)

    def _load_task_schemas(self):

        all_ts = []
        for path in self.config.task_schema_files:
            all_ts.extend(read_YAML_file(path))

        return self.TaskSchemasList.from_json_like(all_ts)

    def shared_data_from_json_like(self, json_like):
        cls_lookup = {
            "command_files": self.CommandFilesList,
            "envs": self.EnvironmentsList,
            "parameters": self.ParametersList,
            "task_schemas": self.TaskSchemasList,
        }
        shared_data = {}
        for k, v in cls_lookup.items():
            shared_data[k] = v.from_json_like(json_like.get(k, {}), is_hashed=True)

        return shared_data


class App(BaseApp):
    """Class to generate an hpcflow application (e.g. MatFlow)"""

    pass



import logging
import os
import platform
import sys
from pathlib import Path
import warnings

import sentry_sdk

from hpcflow.sdk.core.utils import PrettyPrinter


class RunTimeInfo(PrettyPrinter):
    """Get useful run-time information, including the executable name used to
    invoke the CLI, in the case a PyInstaller-built executable was used.

    Attributes
    ----------
    sys_prefix : str
        From `sys.prefix`. If running in a virtual environment, this will point to the
        environment directory. If not running in a virtual environment, this will point to
        the Python installation root.
    sys_base_prefix : str
        From `sys.base_prefix`. This will be equal to `sys_prefix` (`sys.prefix`) if not
        running within a virtual environment. However, if running within a virtual
        environment, this will be the Python installation directory, and `sys_prefix` will
        be equal to the virtual environment directory.
    """

    def __init__(self, name, version, logger):

        is_frozen = getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")
        bundle_dir = (
            sys._MEIPASS if is_frozen else os.path.dirname(os.path.abspath(__file__))
        )

        self.name = name.split(".")[0]  # if name is given as __name__
        self.version = version
        self.is_frozen = is_frozen
        self.working_dir = os.getcwd()
        self.logger = logger

        path_exec = Path(sys.executable)
        path_argv = Path(sys.argv[0])

        if self.is_frozen:
            self.bundle_dir = Path(bundle_dir)
            self.executable_path = path_argv
            self.resolved_executable_path = path_exec
            self.executable_name = self.executable_path.name
            self.resolved_executable_name = self.resolved_executable_path.name
        else:
            self.script_path = path_argv
            self.python_executable_path = path_exec

        self.python_version = platform.python_version()
        self.is_venv = hasattr(sys, "real_prefix") or sys.base_prefix != sys.prefix
        self.is_conda_venv = "CONDA_PREFIX" in os.environ

        self.sys_prefix = getattr(sys, "prefix", None)
        self.sys_base_prefix = getattr(sys, "base_prefix", None)
        self.sys_real_prefix = getattr(sys, "real_prefix", None)
        self.conda_prefix = os.environ.get("CONDA_PREFIX")

        try:
            self.venv_path = self._set_venv_path()
        except ValueError:
            self.venv_path = None

        self.logger.debug(
            f"is_frozen: {self.is_frozen!r}"
            f"{f' ({self.executable_name!r})' if self.is_frozen else ''}"
        )
        self.logger.debug(
            f"is_venv: {self.is_venv!r}"
            f"{f' ({self.sys_prefix!r})' if self.is_venv else ''}"
        )
        self.logger.debug(
            f"is_conda_venv: {self.is_conda_venv!r}"
            f"{f' ({self.conda_prefix!r})' if self.is_conda_venv else ''}"
        )
        if self.is_venv and self.is_conda_venv:
            msg = (
                "Running in a nested virtual environment (conda and non-conda). "
                "Environments may not be re-activate in the same order in associated, "
                "subsequent invocations of hpcflow."
            )
            warnings.warn(msg)

        for k, v in self._get_members().items():
            if k in (
                "is_frozen",
                "is_venv",
                "is_conda_venv",
                "executable_name",
                "python_version",
            ):
                sentry_sdk.set_tag(f"rti.{k}", v)

    def _get_members(self):
        out = {"is_frozen": self.is_frozen, "python_version": self.python_version}
        if self.is_frozen:
            out.update(
                {
                    "executable_name": self.executable_name,
                    "resolved_executable_name": self.resolved_executable_name,
                    "executable_path": self.executable_path,
                    "resolved_executable_path": self.resolved_executable_path,
                }
            )
        else:
            out.update(
                {
                    "script_path": self.script_path,
                    "python_executable_path": self.python_executable_path,
                    "is_venv": self.is_venv,
                    "is_conda_venv": self.is_conda_venv,
                    "sys_prefix": self.sys_prefix,
                    "sys_base_prefix": self.sys_base_prefix,
                    "sys_real_prefix": self.sys_real_prefix,
                    "conda_prefix": self.conda_prefix,
                    "venv_path": self.venv_path,
                }
            )
        out.update({"working_dir": self.working_dir})
        return out

    def __repr__(self):
        out = f"{self.__class__.__name__}("
        for k, v in self._get_members().items():
            out += f"{k}={v!r}"
        return out

    def _set_venv_path(self):
        out = []
        if self.is_venv:
            out.append(self.sys_prefix)
        elif self.is_conda_venv:
            out.append(self.conda_prefix)
        if not out:
            raise ValueError("Not running in a virtual environment!")
        if len(out) == 1:
            return out[0]
        else:
            return out

    def get_activate_env_command(self):
        pass

    def get_deactivate_env_command(self):
        pass



"""Module to define an extensible hpcFlow application class."""
import logging
import os
import sys

_SDK_CONSOLE_LOG_LEVEL = os.environ.get("HPCFLOW_SDK_CONSOLE_LOG_LEVEL", "ERROR")

SDK_logger = logging.getLogger()
SDK_logger.setLevel("DEBUG")

_sh = logging.StreamHandler()
_sh.setFormatter(logging.Formatter("%(levelname)s %(name)s: %(message)s"))
_sh.setLevel(_SDK_CONSOLE_LOG_LEVEL)
SDK_logger.addHandler(_sh)

if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
    import multiprocessing

    multiprocessing.freeze_support()

from hpcflow.sdk.app import App
from hpcflow.sdk.config import ConfigOptions



import copy

import numpy as np

from hpcflow.sdk.core.utils import get_in_container, get_relative_path, set_in_container


PRIMITIVES = (
    int,
    float,
    str,
    type(None),
)


def zarr_encode(data, zarr_group):
    def _zarr_encode(obj, zarr_group, path=None, encoded=None):

        path = path or []
        encoded = encoded or []

        if len(path) > 50:
            raise RuntimeError("I'm in too deep!")

        if isinstance(obj, ZarrEncodable):
            obj = obj.to_dict()
            print(f"obj as dict: {obj}")
            out, encoded = _zarr_encode(
                obj, zarr_group=zarr_group, path=path, encoded=encoded
            )

        elif isinstance(obj, (list, tuple, set)):
            out = []
            for idx, item in enumerate(obj):
                item, encoded = _zarr_encode(item, zarr_group, path + [idx], encoded)
                out.append(item)
            if isinstance(obj, tuple):
                out = tuple(out)
            elif isinstance(obj, set):
                out = set(out)

        elif isinstance(obj, dict):
            out = {}
            for dct_key, dct_val in obj.items():
                dct_val, encoded = _zarr_encode(
                    dct_val, zarr_group, path + [dct_key], encoded
                )
                out.update({dct_key: dct_val})

        elif isinstance(obj, PRIMITIVES):
            out = obj

        elif isinstance(obj, np.ndarray):
            names = [int(i) for i in zarr_group.keys()]
            if not names:
                new_name = "0"
            else:
                new_name = str(max(names) + 1)

            zarr_group.create_dataset(name=new_name, data=obj)
            encoded.append(
                {
                    "path": path,
                    "dataset": new_name,
                }
            )
            out = None

        return out, encoded

    data, encoded = _zarr_encode(data, zarr_group)
    zarr_group.attrs["encoded"] = encoded
    zarr_group.attrs["data"] = data


def zarr_decode(zarr_group, path=None, dataset_copy=False):

    path = path or []

    data = get_in_container(zarr_group.attrs["data"], path)
    data = copy.deepcopy(data)

    for item in zarr_group.attrs["encoded"]:

        try:
            rel_path = get_relative_path(item["path"], path)
        except ValueError:
            continue

        dataset = zarr_group.get(item["dataset"])
        if dataset_copy:
            dataset = dataset[:]

        if rel_path:
            set_in_container(data, rel_path, dataset)
        else:
            data = dataset

    return data


class ZarrEncodable:

    _typ = None

    def to_dict(self):
        if hasattr(self, "__dict__"):
            return dict(self.__dict__)
        elif hasattr(self, "__slots__"):
            return {k: getattr(self, k) for k in self.__slots__}

    def to_zarr(self, zarr_group):
        data = self.to_dict()
        zarr_encode(data, zarr_group)

    @classmethod
    def from_zarr(cls, zarr_group, dataset_copy=False):
        data = zarr_decode(zarr_group, dataset_copy=dataset_copy)
        return cls(**data)



from dataclasses import dataclass
from typing import List, Any

from hpcflow.sdk.core.json_like import JSONLike


@dataclass
class Command(JSONLike):

    command: str
    arguments: List[Any] = None
    stdout: str = None
    stderr: str = None
    stdin: str = None


@dataclass
class CommandArgument:
    """
    Attributes
    ----------
    parts : list of any of str, File, Parameter

    """

    parts: List[Any]



import pytest

from hpcflow.api import hpcflow, WorkflowTemplate, Workflow
from hpcflow.sdk.core.errors import WorkflowNotFoundError


@pytest.fixture
def null_config(tmp_path):
    hpcflow.load_config(config_dir=tmp_path)


def test_make_empty_workflow(null_config, tmp_path):
    Workflow.from_template(WorkflowTemplate(name="w1"), path=tmp_path)


def test_raise_on_missing_workflow(tmp_path):
    with pytest.raises(WorkflowNotFoundError):
        Workflow(tmp_path)



import pytest

from click.testing import CliRunner

from hpcflow import __version__
from hpcflow.api import hpcflow


def test_version():
    runner = CliRunner()
    result = runner.invoke(hpcflow.CLI, args="--version")
    assert result.output.strip() == f"hpcflow, version {__version__}"



from dataclasses import dataclass, field
from typing import List, Optional, Union

from .actions import Action
from .json_like import ChildObjectSpec, JSONLike
from .parameters import (
    Parameter,
    ParameterPropagationMode,
    SchemaInput,
    SchemaOutput,
    SchemaParameter,
)
from .utils import check_valid_py_identifier


@dataclass
class TaskObjective(JSONLike):
    name: str

    def __post_init__(self):
        self.name = check_valid_py_identifier(self.name)


class TaskSchema(JSONLike):

    # TODO: build comprehensive test suite for TaskSchema
    # - decide how schema inputs and outputs are linked to action inputs and outputs?
    # - should the appearance of parameters in the actions determined schema inputs/outputs
    #   - still need to define e.g. defaults, so makes sense to keep inputs/outputs as schema
    #     parameters, and then verify that all action parameters are taken from schema parameters.
    # - should command files be listed as part of the schema? probably, yes.
    _validation_schema = "task_schema_spec_schema.yaml"
    _hash_value = None
    _child_objects = (
        ChildObjectSpec(name="objective", class_name="TaskObjective"),
        ChildObjectSpec(name="inputs", class_name="SchemaInput", is_multiple=True),
        ChildObjectSpec(name="outputs", class_name="SchemaOutput", is_multiple=True),
        ChildObjectSpec(name="actions", class_name="Action", is_multiple=True),
    )

    def __init__(
        self,
        objective: Union[TaskObjective, str],
        actions: List[Action],
        method: Optional[str] = None,
        implementation: Optional[str] = None,
        inputs: Optional[List[Union[Parameter, SchemaInput]]] = None,
        outputs: Optional[List[Union[Parameter, SchemaOutput]]] = None,
        version: Optional[str] = None,
        _hash_value: Optional[str] = None,
    ):
        self.objective = objective
        self.actions = actions
        self.method = method
        self.implementation = implementation
        self.inputs = inputs or []
        self.outputs = outputs or []
        self._hash_value = _hash_value

        self._validate()
        self.version = version

        # if version is not None:  # TODO: this seems fragile
        #     self.assign_versions(
        #         version=version,
        #         app_data_obj_list=self.app.task_schemas
        #         if self.app.is_data_files_loaded
        #         else [],
        #     )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"objective={self.objective.name!r}, "
            f"input_types={self.input_types!r}, "
            f"output_types={self.output_types!r}"
            f")"
        )

    def __eq__(self, other):
        if type(other) is not self.__class__:
            return False
        if (
            self.objective == other.objective
            and self.actions == other.actions
            and self.method == other.method
            and self.implementation == other.implementation
            and self.inputs == other.inputs
            and self.outputs == other.outputs
            and self.version == other.version
            and self._hash_value == other._hash_value
        ):
            return True
        return False

    def _validate(self):

        if isinstance(self.objective, str):
            self.objective = self.app.TaskObjective(self.objective)

        if self.method:
            self.method = check_valid_py_identifier(self.method)
        if self.implementation:
            self.implementation = check_valid_py_identifier(self.implementation)

        # coerce Parameters to SchemaInputs
        for idx, i in enumerate(self.inputs):
            if isinstance(i, Parameter):
                self.inputs[idx] = self.app.SchemaInput(i)

        # coerce Parameters to SchemaOutputs
        for idx, i in enumerate(self.outputs):
            if isinstance(i, Parameter):
                self.outputs[idx] = self.app.SchemaOutput(i)

    @property
    def name(self):
        out = (
            f"{self.objective.name}"
            f"{f'_{self.method}' if self.method else ''}"
            f"{f'_{self.implementation}' if self.implementation else ''}"
        )
        return out

    @property
    def input_types(self):
        return tuple(i.typ for i in self.inputs)

    @property
    def output_types(self):
        return tuple(i.typ for i in self.outputs)

    @property
    def provides_parameters(self):
        return tuple(
            i
            for i in self.inputs + self.outputs
            if i.propagation_mode != ParameterPropagationMode.NEVER
        )

    @classmethod
    def get_by_key(cls, key):
        """Get a config-loaded task schema from a key."""
        return cls.app.task_schemas[key]

    def get_parameter_dependence(self, parameter: SchemaParameter):
        """Find if/where a given parameter is used by the schema's actions."""
        out = {"input_file_writers": [], "commands": []}
        for act_idx, action in enumerate(self.actions):
            deps = action.get_parameter_dependence(parameter)
            for key in out:
                out[key].extend((act_idx, i) for i in deps[key])
        return out

    def get_key(self):
        return (str(self.objective), self.method, self.implementation)
